# Longbow is Copyright (C) of James T Gebbie-Rayet and Gareth B Shannon 2015.
#
# This file is part of the Longbow software which was developed as part of
# the HECBioSim project (http://www.hecbiosim.ac.uk/).
#
# HECBioSim facilitates and supports high-end computing within the
# UK biomolecular simulation community on resources such as ARCHER.
#
# Longbow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Longbow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Longbow.  If not, see <http://www.gnu.org/licenses/>.

"""
The applications module contains methods for processing the aspect of jobs
which relate to external applications (such as an MD package). The following
methods can be found within this module:

testapp()
    This method will make an attempt to check that the application executables
    required to run a job/s is present on the specified host/s. This method is
    capable of using the module system.

processjobs()
    This method will process information that is given as an intended target to
    be passed on to the executable at run time. It will check that required
    parameters (provided the respective plug-in is configured correctly) have
    been supplied, and that all files and their dependencies (again provided
    that the respective plug-in is configured for this) exist on disk.
"""

import logging
import os
from random import randint

# Depending on how longbow is installed/utilised the import will be slightly
# different, this should handle both cases.
try:

    EX = __import__("corelibs.exceptions", fromlist=[''])
    SHELLWRAPPERS = __import__("corelibs.shellwrappers", fromlist=[''])
    APPS = __import__("plugins.apps", fromlist=[''])

except ImportError:

    EX = __import__("Longbow.corelibs.exceptions", fromlist=[''])
    SHELLWRAPPERS = __import__("Longbow.corelibs.shellwrappers", fromlist=[''])
    APPS = __import__("Longbow.plugins.apps", fromlist=[''])

LOGGER = logging.getLogger("Longbow")


def testapp(hosts, jobs):

    """
    This method will make an attempt to check that the application executables
    required to run a job/s is present on the specified host/s. This method is
    capable of using the module system.

    Required arguments are:

    hosts (dictionary) - The Longbow hosts data structure, see configuration.py
                         for more information about the format of this
                         structure.

    jobs (dictionary) - The Longbow jobs data structure, see configuration.py
                        for more information about the format of this
                        structure.
    """

    checked = {}

    LOGGER.info("Testing the executables defined for each job.")

    for job in jobs:

        resource = jobs[job]["resource"]
        executable = jobs[job]["executable"]

        # If we haven't checked this resource then it is likely not in the dict
        if resource not in checked:

            checked[resource] = []

        # Now check if we have tested this exec already.
        if executable not in checked[resource]:

            # If not then add it to the list now.
            checked[resource].extend([executable])

            LOGGER.info(
                "Checking executable '{0}' on '{1}'"
                .format(executable, resource))

            cmd = []

            if jobs[job]["modules"] is "":

                LOGGER.debug("Checking without modules.")

            else:
                LOGGER.debug("Checking with modules.")

                for module in jobs[job]["modules"].split(","):

                    module = module.replace(" ", "")
                    cmd.extend(["module load " + module + "\n"])

            cmd.extend(["which " + executable])

            try:

                SHELLWRAPPERS.sendtossh(hosts[resource], cmd)
                LOGGER.info("Executable check - passed.")

            except EX.SSHError:

                LOGGER.error("Executable check - failed.")
                raise


def processjobs(jobs):

    """
    This method will process information that is given as an intended target to
    be passed on to the executable at run time. It will check that required
    parameters (provided the respective plug-in is configured correctly) have
    been supplied, and that all files and their dependencies (again provided
    that the respective plug-in is configured for this) exist on disk.

    Required arguments are:

    jobs (dictionary) - The Longbow jobs data structure, see configuration.py
                        for more information about the format of this
                        structure.
    """

    LOGGER.info("Processing job/s and detecting files that require upload.")

    # Get dictionary of executables and their required flags from plug-ins.
    tmp = getattr(APPS, "DEFMODULES")

    # Process each job.
    for job in jobs:

        # Initialise some basic parameters.
        executable = jobs[job]["executable"]
        app = tmp[executable]
        args = jobs[job]["executableargs"]
        filelist = []

        # Append a hash to the job directory to avoid directory clashes.
        destdir = job + ''.join(["%s" % randint(0, 9) for _ in range(0, 5)])

        jobs[job]["destdir"] = os.path.join(jobs[job]["destdir"], destdir)

        LOGGER.debug("Job '{0}' will be run in the '{1}' directory on the "
                     "remote resource.".format(job, jobs[job]["destdir"]))

        LOGGER.debug("Command-line arguments for job '{0}'are '{1}'"
                     .format(job, args))

        # Check for any files that are located outside the work directory or
        # absolute paths.
        for item in args:

            if item.count(os.path.pardir) > 0 or os.path.isabs(item):

                raise EX.RequiredinputError(
                    "In job '{0}' input files are being provided with "
                    "absolute paths or from directories above localworkdir. "
                    "This is not supported".format(job))

        # Base path to local job directory.
        cwd = jobs[job]["localworkdir"]

        # If we have multiple jobs.
        if len(jobs) > 1:

            # Add the job name to the path.
            cwd = os.path.join(cwd, job)
            jobs[job]["localworkdir"] = cwd

        # Check that the directory exists.
        if os.path.isdir(cwd) is False:

            # If not, this is bad.
            raise EX.DirectorynotfoundError(
                "The local job directory '{0}' cannot be found for job '{1}'"
                .format(cwd, job))

        # Detect command line substitutions. Make a copy of the commandline
        # arguments as text enforcing substitutions will be removed for most
        # of the code
        initargs = list(args)
        substte = {}

        try:

            substte = getattr(APPS, app.lower()).sub_dict(args)

        except AttributeError:

            pass

        # Process the command line arguments, detecting any missing flags,
        # files or dependencies within the files.

        # Get required flags.
        req_flags = getattr(APPS, "EXECFLAGS")[executable]
        found_flags = []

        # Start with the case where the command line would be provided like
        # exec < input.file or exec input.file.
        if "?" in req_flags:

            # Check if the first item is the "<".
            if args[0] is "<":

                # Check the length of the command line.
                if len(args) > 1:

                    # If 'replicates' == 1 then we will only check one file,
                    # else we will proceed to check files in all replicates.
                    for i in range(1, int(jobs[job]["replicates"]) + 1):

                        # If we do only have a single job then file path should
                        # be
                        if int(jobs[job]["replicates"]) == 1:

                            # For this type of job the file should be at [1].
                            fileitem = args[1]

                        # Otherwise we have a replicate job so we should amend
                        # the paths.
                        else:

                            # We should check that the replicate directory
                            # structure exists. If it doesn't then create it.
                            if os.path.isdir(os.path.join(
                                    cwd, "rep" + str(i))) is False:

                                os.mkdir(os.path.join(cwd, "rep" + str(i)))

                            # Add the repX dir to the file list as rsync will
                            # not create them.
                            if ("rep" + str(i)) not in filelist:

                                filelist.append("rep" + str(i))

                            # If we have a replicate job then we should check
                            # if the file resides within ./rep{i} or if it is
                            # a global (common to each replicate) file.
                            if os.path.isfile(
                                    os.path.join(
                                        cwd, "rep" + str(i), args[1])):

                                fileitem = os.path.join("rep" + str(i),
                                                        args[1])

                            # Otherwise do we have a file in cwd
                            elif os.path.isfile(os.path.join(cwd, args[1])):

                                fileitem = args[1]

                                # Also update the command line to reflect a
                                # global file.
                                if args[1] in initargs:

                                    initargs[initargs.index(args[1])] = \
                                        os.path.join("../", args[1])

                        # If the next argument along is a valid file.
                        if os.path.isfile(os.path.join(cwd, fileitem)):

                            # Then mark the flag as found.
                            found_flags.append("?")

                            # Search input file for any file dependencies that
                            # don't exist.
                            try:

                                getattr(
                                    APPS, app.lower()).file_parser(
                                        fileitem, cwd, filelist, substte)

                            except AttributeError:

                                if fileitem not in filelist:
                                    filelist.append(fileitem)

                # Looks like the command line is too short to contain the
                # input file so raise an exception.
                else:

                    raise EX.RequiredinputError(
                        "In job '{0}' it appears that the input file is "
                        "missing, check your command line is of the form "
                        "longbow [longbow args] executable '<' "
                        "[executable args]".format(job))

            # Some programs are run with the format "executable input.file"
            elif args[0] is not "<":

                # Lets make sure that we actually have something to load.
                if len(args) > 0:

                    # If 'replicates' == 1 then we will only check one file,
                    # else we will proceed to check files in all replicates.
                    for i in range(1, int(jobs[job]["replicates"]) + 1):

                        # If we do only have a single job then file path should
                        # be
                        if int(jobs[job]["replicates"]) == 1:

                            # For this type of job the file should be at [0].
                            fileitem = args[0]

                        # Otherwise we have a replicate job so we should amend
                        # the paths.
                        else:

                            # We should check that the replicate directory
                            # structure exists.
                            if os.path.isdir(os.path.join(
                                    cwd, "rep" + str(i))) is False:

                                os.mkdir(os.path.join(cwd, "rep" + str(i)))

                            # Add the repX dir to the file list as rsync will
                            # not create them.
                            if ("rep" + str(i)) not in filelist:

                                filelist.append("rep" + str(i))

                            # If we have a replicate job then we should check
                            # if the file resides within ./rep{i} or if it is
                            # a global (common to each replicate) file.
                            if os.path.isfile(
                                    os.path.join(
                                        cwd, "rep" + str(i), args[0])):

                                fileitem = os.path.join("rep" + str(i),
                                                        args[0])

                            # Otherwise do we have a file in cwd
                            elif os.path.isfile(os.path.join(cwd, args[0])):

                                fileitem = args[0]

                                # Also update the command line to reflect a
                                # global file.
                                if args[0] in initargs:

                                    initargs[initargs.index(args[0])] = \
                                        os.path.join("../", args[0])

                        # If we have something to load then check that it is a
                        # valid file
                        if os.path.isfile(os.path.join(cwd, fileitem)):

                            # Then mark the flag as found.
                            found_flags.append("?")

                            # Search input file for any file dependencies that
                            # don't exist.
                            try:

                                getattr(
                                    APPS, app.lower()).file_parser(
                                        fileitem, cwd, filelist, substte)

                            except AttributeError:

                                if fileitem not in filelist:

                                    filelist.append(fileitem)

                # Looks like the command line is too short to contain the
                # input file so raise an exception.
                else:

                    raise EX.RequiredinputError(
                        "In job '{0}' it appears that the input file "
                        "is missing, check your command line is of "
                        "the form: "
                        "longbow [longbow args] executable '<' "
                        "[executable args]".format(job))

        # Otherwise we have a more conventional command line of the form:
        # exec -a arg1 -b arg2 --foo bar
        else:

            # Run through each one.
            for item in args:

                fileitem = ""

                # If we have a flag (starting with '-') and it is in the list
                # of required flags.
                if item[0] is "-" and item in req_flags:

                    # Mark the flag as found
                    found_flags.append(item)

                # If we have a flag that is not in the list of required flags.
                elif item[0] is "-" and item not in req_flags:

                    # Check then if this flag belongs to a case where it is
                    # either one or the other flag has to be provided (ie ones
                    # listed as -foo || bar) such as gromacs -s or -deffnm)
                    for flag in req_flags:

                        # If it is the special case
                        if "||" in flag:

                            # Split them into separate args.
                            flags = flag.split(" || ")

                            # Now check if flag is present
                            if item in flags:

                                # Mark it as found if it hasn't been already.
                                if flag not in found_flags:
                                    found_flags.append(flag)

                # Otherwise it could just be a file or a parameter.
                else:

                    # If 'replicates' == 1 then we will only check one file,
                    # else we will proceed to check files in all replicates.
                    for i in range(1, int(jobs[job]["replicates"]) + 1):

                        tmpitem = ""

                        # If we do only have a single job then file path should
                        # be
                        if int(jobs[job]["replicates"]) == 1:

                            if os.path.isfile(os.path.join(cwd, item)):

                                fileitem = item

                            # Hook for checking plugin specific file naming
                            # scenarios (eg. gromacs -deffnm test actually
                            # referring to test.tpr).
                            else:

                                try:

                                    fileitem = getattr(
                                        APPS, app.lower()).defaultfilename(
                                            cwd, item)

                                except AttributeError:

                                    pass

                        # Otherwise we have a replicate job so we should amend
                        # the paths.
                        else:

                            # We should check that the replicate directory
                            # structure exists.
                            if os.path.isdir(os.path.join(
                                    cwd, "rep" + str(i))) is False:

                                os.mkdir(os.path.join(cwd, "rep" + str(i)))

                            # Add the repX dir to the file list as rsync will
                            # not create them.
                            if ("rep" + str(i)) not in filelist:

                                filelist.append("rep" + str(i))

                            # If we have a replicate job then we should check
                            # if the file resides within ./rep{i} or if it is
                            # a global (common to each replicate) file.
                            if os.path.isfile(
                                    os.path.join(cwd, "rep" + str(i), item)):

                                # Set the file path
                                fileitem = os.path.join("rep" + str(i), item)

                            # Otherwise do we have a file here.
                            elif os.path.isfile(os.path.join(cwd, item)):

                                # If we do then set file path.
                                fileitem = item

                                # Also update the command line to reflect a
                                # global file.
                                if item in initargs:

                                    initargs[initargs.index(item)] = \
                                        os.path.join("../", item)

                            # Hook for checking plugin specific file naming
                            # scenarios (eg. gromacs -deffnm test actually
                            # referring to test.tpr).
                            else:

                                try:

                                    tmpitem = getattr(
                                        APPS, app.lower()).defaultfilename(
                                            cwd, os.path.join(
                                                "rep" + str(i) + item))

                                except AttributeError:

                                    pass

                                # If we have a positive check then file found
                                # in rep$i directories.
                                if tmpitem is not "":

                                    fileitem = tmpitem

                                # Otherwise check for global one.
                                else:

                                    try:

                                        fileitem = getattr(
                                            APPS, app.lower()).defaultfilename(
                                                cwd, item)

                                        # Also update the command line to
                                        # reflect a global file.
                                        if item in initargs:

                                            initargs[initargs.index(item)] = \
                                                os.path.join("../", item)

                                    except AttributeError:

                                        pass

                        # Check if there are any dependencies inside the file
                        # otherwise if this is not supported then just append
                        # for upload.
                        if os.path.isfile(os.path.join(cwd, fileitem)):

                            # Search input file for any file dependencies
                            # that don't exist.
                            try:

                                getattr(
                                    APPS, app.lower()).file_parser(
                                        fileitem, cwd, filelist, substte)

                            except AttributeError:

                                if fileitem not in filelist:

                                    filelist.append(fileitem)

            # Final check for if any required flags are missing.
            flags = list(set(req_flags) - set(found_flags))

            # If there are any missing still then tell the user.
            if len(flags) is not 0:

                raise EX.RequiredinputError(
                    "In job '{0}' there are missing flags on the command line "
                    "'{1}'. See user documentation for plug-in '{2}'".format(
                        job, flags, getattr(APPS, "DEFMODULES")[executable]))

        # Setup the rysnc upload masks.
        if jobs[job]["upload-include"] is "":

            jobs[job]["upload-include"] = (", ".join(filelist))

        else:

            jobs[job]["upload-include"] = (
                jobs[job]["upload-include"] + ", " + ", ".join(filelist))

        jobs[job]["upload-exclude"] = "*"

        # Replace the input command line with the execution command line.
        # initargs is a copy of the original args before text enforcing
        # substitutions was removed
        jobs[job]["executableargs"] = executable + " " + " ".join(initargs)

        LOGGER.info("For job '{0}' - execution string: {1}".format(
            job, jobs[job]["executableargs"]))

    LOGGER.info("Processing jobs - complete.")



# Longbow is Copyright (C) of James T Gebbie-Rayet and Gareth B Shannon 2015.
#
# This file is part of the Longbow software which was developed as part of
# the HECBioSim project (http://www.hecbiosim.ac.uk/).
#
# HECBioSim facilitates and supports high-end computing within the
# UK biomolecular simulation community on resources such as ARCHER.
#
# Longbow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Longbow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Longbow.  If not, see <http://www.gnu.org/licenses/>.

"""
The staging module provides methods for processing the transfer of files
between the local host and the remote host job directories.

The following methods are contained within this module:

stage_upstream()
    A method for staging files for each job to the target HPC host. The
    underlying utility behind this transfer is rsync, thus it is possible
    to supply rsync file masks to blacklist unwanted large files. By default
    rsync is configured to transfer blockwise and only transfer the
    newest/changed blocks, this saves a lot of time during persistant staging.

stage_downstream()
    A method for staging files for each job to from target HPC host. The
    underlying utility behind this transfer is rsync, thus it is possible
    to supply rsync file masks to blacklist unwanted large files. By default
    rsync is configured to transfer blockwise and only transfer the
    newest/changed blocks, this saves a lot of time during persistant staging.

cleanup()
    A method for cleaning up the working directory on the HPC host, this method
    will only delete job directories that are valid for the given Longbow
    instance, thus avoid data loss.
"""

import logging

try:

    EX = __import__("corelibs.exceptions", fromlist=[''])
    SHELLWRAPPERS = __import__("corelibs.shellwrappers", fromlist=[''])

except ImportError:

    EX = __import__("Longbow.corelibs.exceptions", fromlist=[''])
    SHELLWRAPPERS = __import__("Longbow.corelibs.shellwrappers", fromlist=[''])

LOGGER = logging.getLogger("Longbow")


def stage_upstream(hosts, jobs):

    """
    A method for staging files for each job to the target HPC host. The
    underlying utility behind this transfer is rsync, thus it is possible
    to supply rsync file masks to blacklist unwanted large files. By default
    rsync is configured to transfer blockwise and only transfer the
    newest/changed blocks, this saves a lot of time during persistant staging.

    Required arguments are:

    hosts (dictionary) - The Longbow hosts data structure, see configuration.py
                         for more information about the format of this
                         structure.

    jobs (dictionary) - The Longbow jobs data structure, see configuration.py
                        for more information about the format of this
                        structure.
    """

    LOGGER.info("Staging files for job/s.")

    for job in jobs:

        host = hosts[jobs[job]["resource"]]
        src = jobs[job]["localworkdir"]
        dst = jobs[job]["destdir"]

        LOGGER.info("Transfering files for job '{0}' to host '{1}'"
                    .format(job, jobs[job]["resource"]))

        cmd = []
        cmd.extend(["mkdir -p " + dst + "\n"])

        try:

            SHELLWRAPPERS.sendtossh(host, cmd)

            LOGGER.info("mkdir -p '{0}' passed.".format(dst))

        except EX.SSHError:

            LOGGER.error("mkdir -p '{0}' failed.".format(dst))
            raise

        # Transfer files upstream.
        try:

            SHELLWRAPPERS.upload(
                host, src + "/", dst,
                jobs[job]["upload-include"],
                jobs[job]["upload-exclude"])

        except EX.RsyncError:

            raise EX.StagingError("Could not stage file '{0}' upstream"
                                  .format(src))

    LOGGER.info("Staging files upstream - complete.")


def stage_downstream(hosts, jobs, jobname):

    """
    A method for staging files for each job to from target HPC host. The
    underlying utility behind this transfer is rsync, thus it is possible
    to supply rsync file masks to blacklist unwanted large files. By default
    rsync is configured to transfer blockwise and only transfer the
    newest/changed blocks, this saves a lot of time during persistant staging.

    Required arguments are:

    hosts (dictionary) - The Longbow hosts data structure, see configuration.py
                         for more information about the format of this
                         structure.

    jobs (dictionary) - The Longbow jobs data structure, see configuration.py
                        for more information about the format of this
                        structure.

    jobname (string) - A string that matches the name of a job to be staged.
                       (primary key in jobs dict)
    """

    LOGGER.info("For job '{0}' staging files downstream.".format(jobname))

    host = hosts[jobs[jobname]["resource"]]
    src = jobs[jobname]["destdir"] + "/"
    dst = jobs[jobname]["localworkdir"]

    # Download the whole directory with rsync.
    try:

        SHELLWRAPPERS.download(
            host, src, dst,
            jobs[jobname]["download-include"],
            jobs[jobname]["download-exclude"])

    except (EX.SCPError, EX.RsyncError):

        raise EX.StagingError("Could not download file '{0}' "
                              "to location '{1}'".format(src, dst))

    LOGGER.info("Staging complete.")


def cleanup(hosts, jobs):

    """
    A method for cleaning up the working directory on the HPC host, this method
    will only delete job directories that are valid for the given Longbow
    instance, thus avoid data loss.

    Required arguments are:

    hosts (dictionary) - The Longbow hosts data structure, see configuration.py
                         for more information about the format of this
                         structure.

    jobs (dictionary) - The Longbow jobs data structure, see configuration.py
                        for more information about the format of this
                        structure.
    """

    LOGGER.info("Cleaning up the work directories.")

    for job in jobs:

        try:

            host = hosts[jobs[job]["resource"]]
            path = jobs[job]["destdir"]

            SHELLWRAPPERS.remotelist(host, path)

            if path != hosts[jobs[job]["resource"]]["remoteworkdir"]:
                LOGGER.info("Deleting directory for job '{0}' - '{1}'"
                            .format(job, path))

                SHELLWRAPPERS.remotedelete(hosts[jobs[job]["resource"]], path)

            else:
                raise EX.RemoteworkdirError("Subdirectory of remoteworkdir" +
                                            " not yet created")

        except EX.RemoteworkdirError:

            LOGGER.debug("For job '{0}', cleanup not required because"
                         " the '{1}xxxxx' subdirectory of {2} in which the job"
                         " would have run has not not yet been created on the"
                         " remote resource.".format(job,
                                                    job,
                                                    host["remoteworkdir"]))

        except EX.RemotelistError:

            # Directory doesn't exist.
            LOGGER.debug("Directory on path '{0}' does not exist - skipping."
                         .format(path))

        except KeyError:

            LOGGER.debug("For job '{0}', cleanup not required - skipping."
                         .format(job))

        except EX.RemotedeleteError:

            LOGGER.debug("For job '{0}', cannot delete directory '{1}' - "
                         "skipping.".format(job, path))

    LOGGER.info("Cleaning up complete.")



# Longbow is Copyright (C) of James T Gebbie-Rayet and Gareth B Shannon 2015.
#
# This file is part of the Longbow software which was developed as part of
# the HECBioSim project (http://www.hecbiosim.ac.uk/).
#
# HECBioSim facilitates and supports high-end computing within the
# UK biomolecular simulation community on resources such as ARCHER.
#
# Longbow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Longbow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Longbow.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains methods for loading and saving to Longbow (ini)
configuration files in addition to methods for extracting the resultant
information into Longbow data structures. The templates for these data
structures and their forms are also declared within this module.

The following data structures can be found:

HOSTTEMPLATE
    The template of the hosts data structure. The Longbow API will assume
    that variables listed here are to be found in this structure.

HOSTREQUIRED
    A dictionary to mark parameters that are required to be initialised during
    configuration and their error messages for cases when missing.

JOBTEMPLATE
    The template of the job data structure. The Longbow API will assume
    that variables listed here are to be found in this structure.

JOBREQUIRED
    A dictionary to mark parameters that are required to be initialised during
    configuration and their error messages for cases when missing.

The following methods can be found:

processjobs()
    Method for processing the raw configuration structures loaded from the
    configuration files into Longbow friendly configuration structures.
    This is where the parameter hierarchy is applied.

loadconfigs()
    Method for loading and extracting data from the Longbow configuration
    files.

saveconfigs()
    Method for saving data to Longbow configuration files, this method will
    honour comments and simply amend the file structure with new or changed
    data.
"""

import logging
import re

# Depending on how Longbow is installed/utilised the import will be slightly
# different, this should handle both cases.
try:

    EX = __import__("corelibs.exceptions", fromlist=[''])
    APPS = __import__("plugins.apps", fromlist=[''])

except ImportError:

    EX = __import__("Longbow.corelibs.exceptions", fromlist=[''])
    APPS = __import__("Longbow.plugins.apps", fromlist=[''])


LOGGER = logging.getLogger("Longbow")

HOSTTEMPLATE = {
    "account": "",
    "accountflag": "",
    "corespernode": "24",
    "handler": "",
    "host": "",
    "port": "22",
    "remoteworkdir": "",
    "scheduler": "",
    "user": ""
}

HOSTREQUIRED = {
    "host": "The parameter 'hosts' has not been set in your configuration "
            "file, the most natural place is to set this in your hosts.conf, "
            "the path is normally what you would supply to SSH (the part "
            "after the '@').",
    "user": "The paramter 'users' has not been set in your configuration "
            "file, the most natural place is to set this in your hosts.conf, "
            "the user is the same as what you would normally supply to SSH "
            "(the part before the '@').",
    "remoteworkdir": "The parameter 'remoteworkdir' has not been set in your "
                     "configuration file, the most natural place to set this "
                     "is in the host.conf if your working directory never "
                     "changes, or in the job.conf if you are changing this on "
                     "a per job basis."
}

JOBTEMPLATE = {
    "cores": "24",
    "cluster": "",
    "download-exclude": "",
    "download-include": "",
    "executable": "",
    "executableargs": "",
    "frequency": "300",
    "localworkdir": "",
    "modules": "",
    "maxtime": "24:00",
    "memory": "",
    "queue": "",
    "resource": "",
    "replicates": "1",
    "upload-exclude": "",
    "upload-include": ""
}

JOBREQUIRED = {
    "executableargs": "Command-line arguments could not be detected properly "
                      "on the command-line or in a configuration file. If "
                      "your application requires input of the form "
                      "'executable < input_file' then make sure that you put "
                      "the '<' in quotation marks.",
    "executable": "An executable has not been specified on the command-line "
                  "or in a configuration file.",
    "replicates": "The parameter 'replicates' is required to be set, this "
                  "parameter has an internal default for when not specified "
                  "so something may have gone wrong when specifying a value "
                  "for it in the configuration file."
}


def processconfigs(hostfile, jobfile, cwd, params):

    """
    Method for processing the raw configuration structures loaded from the
    configuration files into Longbow friendly configuration structures.
    This is where the parameter hierarchy is applied.

    Required arguments are:

    hostfile (string): This should be an absolute path to a configuration
                       file, this parameter is required.

    jobfile (string): This should be an absolute path to a configuration
                      file, this parameter is optional (empty string if not
                      needed).

    args (string): This should be an absolute path to a configuration
                       file, this parameter is required.

    cwd (string): This should be an absolute path to a configuration
                       file, this parameter is required.

    executable (string): This should be an absolute path to a configuration
                       file, this parameter is required.

    params (dictionary): This should be an absolute path to a configuration
                       file, this parameter is required.

    Return parameters are:

    hosts (dictionary): A fully processed Longbow hosts data structure.

    jobs (dictionary) A fully processed Longbow jobs data structure.
    """

    # Define our main data structures.
    hosts = {}
    jobs = {}

    # Define a dictionary of module defaults based on the plug-in names and
    # executables.
    modules = getattr(APPS, "DEFMODULES")
    modules[""] = ""

    # Try and load the host file.
    try:

        _, hostsections, hostdata = loadconfigs(hostfile)

    except EX.ConfigurationError:

        raise

    # If we have been given a job file then try and load it.
    if jobfile is not "":

        try:

            _, _, jobdata = loadconfigs(jobfile)

        except EX.ConfigurationError:

            raise

    # If there is no file then the job structure will be built up from things
    # we know from the hosts, the command line and defaults. In this case we
    # would only ever have 1 job.
    else:

        jobdata = {}

        # Did the user supply a jobname on the command line, if not default it.
        if params["jobname"] is not "":

            jobname = params["jobname"]

        else:
            jobname = "LongbowJob"

        jobdata[jobname] = JOBTEMPLATE.copy()

        # It is important that this one is empty so not to accidentally
        # interfere with prioritisation in the next step (saves on duplicating
        # data structures just to have one empty and one with default values).
        for item in jobdata[jobname]:

            jobdata[jobname][item] = ""

    # Process the basic host configuration
    for host in hostdata:

        # Create a base host structure along with known defaults.
        hosts[host] = HOSTTEMPLATE.copy()

        # At this stage just grab all parameters that belong in the hosts.
        for item in hosts[host]:

            # We don't want to overwrite internal defaults.
            try:

                if hostdata[host][item] is not "":

                    hosts[host][item] = hostdata[host][item]

            except KeyError:

                pass

    # Process the basic job configuration.
    for job in jobdata:

        # Create a base job structure along with known defaults.
        jobs[job] = JOBTEMPLATE.copy()

        for item in jobs[job]:

            # At this stage just grab all parameters that belong in the jobs,
            # which are not empty to avoid overwriting defaults.
            try:

                if jobdata[job][item] is not "":

                    jobs[job][item] = jobdata[job][item]

            except KeyError:

                pass

    # Now process some overrides, the basic order of precedence is;
    # Command line > job config file > host config file > internal defaults.

    # Before we go further lets check that the job/s have been assigned a host.
    # This is important for checking for values provided for jobs in the host
    # conf.
    for job in jobs:

        # If not then can we find one to use?
        if jobs[job]["resource"] is "":

            # First check the params from the command line.
            if params["resource"] is not "":

                # If this host has been set up then use it.
                if params["resource"] in hostsections:

                    jobs[job]["resource"] = params["resource"]

                # Otherwise we have a problem, so tell the user.
                else:

                    raise EX.CommandlineargsError(
                        "The resource '{0}' that was given on the command line"
                        " has not been configured in the host.conf. The hosts "
                        "available are '{0}'".format(
                            params["resource"], hostsections))

            # Otherwise lets try and use the top host in the list from
            # host.conf. This should never be an empty list since the
            # parser would provide an error on load.
            else:

                jobs[job]["resource"] = hostsections[0]

        # Otherwise run some validation.
        else:

            # Does the host exist?
            if jobs[job]["resource"] not in hostsections:

                # If not tell the user.
                raise EX.CommandlineargsError(
                    "The resource '{0}' that was given in the job config file"
                    " has not been configured in the host.conf. The hosts "
                    "available are '{0}'".format(
                        params["resource"], hostsections))

    # Now we can go ahead and process the overrides. For this it is probably
    # best to use the jobs data as the main source, this way we will only deal
    # with hosts that are being used.
    for job in jobs:

        resource = jobs[job]["resource"]

        # Host overrides first.

        # For the host that is provided for this job, is there anything that is
        # overriding in the job configuration file?
        for item in jobdata[job]:

            # We can check this like this.
            if item in hosts[resource] and jobdata[job][item] is not "":

                # Then override.
                hosts[resource][item] = jobdata[job][item]

        # For the host that is provided in this job, is there anything that is
        # overriding on the command line?
        for item in params:

            # We can check this like this.
            if item in hosts[resource] and params[item] is not "":

                # Then override.
                hosts[resource][item] = params[item]

        # Job overrides next.

        # Now the overrides from the hosts.conf are of lower priority than
        # those on the command line or any data that was supplied in the
        # job.conf but higher than any internal defaults.
        for item in hostdata[resource]:

            # We have already handled this.
            if item is not "resource":

                # Does the item match anything in the job?
                if item in jobs[job]:

                    # Then we have an override, but check priority first.
                    # Make sure item didn't come from higher priority job.conf?
                    if item not in jobdata[job] or jobdata[job][item] is "":

                        # Override.
                        jobs[job][item] = hostdata[resource][item]

        # For the currently selected job, are there any overrides that have
        # been given on the command line?
        for item in params:

            # We have already handled this.
            if item is not "resource":

                # We can check this like this.
                if item in jobs[job] and params[item] is not "":

                    # Then override.
                    jobs[job][item] = params[item]

    # Check parameters that are required for running jobs are provided.
    # Here we will only do validation on hosts that are referenced in jobs,
    # this should cut down on annoyances to the user.
    for job in jobs:

        resource = jobs[job]["resource"]

        # Validate for host referenced within job.
        for validationitem in HOSTREQUIRED:

            # If no value has been set.
            if hosts[resource][validationitem] is "":

                # Throw an exception.
                raise EX.ConfigurationError(HOSTREQUIRED[validationitem])

        # Validate required parameters have been set.
        for validationitem in JOBREQUIRED:

            # If no value has been set.
            if jobs[job][validationitem] is "":

                # Throw an exception.
                raise EX.ConfigurationError(JOBREQUIRED[validationitem])

    # Some final initialisation.
    for job in jobs:

        # If the local working directory has not been set, then default to cwd
        if jobs[job]["localworkdir"] is "":

            jobs[job]["localworkdir"] = cwd

        # Fix for python 3 where basestring is now str.
        try:

            # If the exec arguments are in string form, split to list.
            if isinstance(jobs[job]["executableargs"], basestring):

                jobs[job]["executableargs"] = (
                    jobs[job]["executableargs"].split())

        except NameError:

            # If the exec arguments are in string form, split to list.
            if isinstance(jobs[job]["executableargs"], str):

                jobs[job]["executableargs"] = (
                    jobs[job]["executableargs"].split())

        # If modules hasn't been set then try and use a default.
        if jobs[job]["modules"] is "":

            jobs[job]["modules"] = modules[jobs[job]["executable"]]

        # Give each job a remote base path, a random hash will be added to this
        # during job processing
        jobs[job]["destdir"] = hosts[jobs[job]["resource"]]["remoteworkdir"]

    return hosts, jobs


def loadconfigs(configfile):

    """
    Method to load an ini file. Files of this format contain the following
    mark-up structure.

    Sections of a file are marked using square brackets
    Section then contain option statements of the form "param = value"
    Comments are marked using hashes.

    An example of such a file would be:

    [section1]
    # this is the first option
    option1 = value1
    # this is the second option
    option2 = value2

    [section2]
    # this is the first option
    option1 = value1
    # this is the second option
    option2 = value2

    [section3]
    # this is the first option
    option1 = value1
    # this is the second option
    option2 = value2

    This method performs basic error handling to do with the structure of the
    ini file only. All error handling specific to Longbow should be performed
    elsewhere.

    Required arguments are:

    configfile (string): This should be an absolute path to a configuration
                         file.

    Return parameters are:

    contents (list): This is the raw file structure where each line is an item
                     in the list.

    sections (list): This is a list of section headers in the data (preserves
                     order).

    data (dict of dicts): This is a structure containing the data loaded from
                          the file, a dictionary is created for each heading in
                          the ini file. Then the parameters and values under
                          each heading will form a dictionary within the
                          corresponding heading section (dictionary of
                          dictionaries).
    """

    LOGGER.info("Loading configuration information from file '{0}'"
                .format(configfile))

    sections = []
    params = {}

    # Open configuration file.
    try:

        tmp = open(configfile, "r")

        # Grab all of the information from the file.
        contents = tmp.readlines()

    except IOError:

        raise EX.ConfigurationError(
            "Can't read the configurations from '{0}'".format(configfile))

    # Strip out the newline chars.
    contents = [line.strip("\n") for line in contents]

    # Close configuration file (don't need it anymore).
    tmp.close()

    # Pull out sections into list.
    for item in contents:

        # Don't care about blank lines.
        if len(item) > 0:

            try:
                # Find section markers
                if item[0] == "[" and item[len(item)-1] == "]":

                    # Remove the square bracket section markers.
                    section = "".join(a for a in item if a not in "[]")

                    # Add to list of sections.
                    sections.append(section)

                    # Create a new section in the data structure.
                    params[section] = {}

                # Find comment markers.
                elif item[0] is "#":

                    # Ignore comments.
                    pass

                # Anything else must be option data.
                else:

                    # Option is in the format key = param. Added regular
                    # expression so that if user writes with/without spaces
                    # then we can still extract the configuration info.
                    key, value = re.split(" = |= | =|=", item)

                    # Store the keys and values in the data structure.
                    params[section][key] = value

            except NameError:

                # Issue warning.
                pass

    # Check if there are zero sections.
    if len(sections) is 0:

        raise EX.ConfigurationError(
            "Error no sections are defined in configuration file '{0}'"
            .format(configfile))

    # Check for sections with zero options.
    for section in sections:

        if len(params[section]) is 0:

            raise EX.ConfigurationError(
                "Error section '{0}' contains no parameter definitions using "
                "configuration file '{1}'".format(section, configfile))

    return contents, sections, params


def saveconfigs(configfile, params):

    """
    Method to saving to an ini file. Files of this format contain the
    following mark-up structure.

    Sections of a file are marked using square brackets
    Section then contain option statements of the form "param = value"
    Comments are marked using hashes.

    An example of such a file would be:

    [section1]
    # this is the first option
    option1 = value1
    # this is the second option
    option2 = value2

    [section2]
    # this is the first option
    option1 = value1
    # this is the second option
    option2 = value2

    [section3]
    # this is the first option
    option1 = value1
    # this is the second option
    option2 = value2

    This method is comment safe, this was a major downfall of the standard
    python parser as it would wipe out comments that a user would include.

    Required arguments are:

    configfile (string): This should be an absolute path to a configuration
                         file.

    params (dictionary): This should contain the data structure that should
                         be saved (typically hosts of job configs structure).
    """

    LOGGER.info("Saving configuration information to file '{0}'"
                .format(configfile))

    keydiff = {}
    valuediff = {}

    # Load up the original file including comment structure (list).
    contents, _, oldparams = loadconfigs(configfile)

    # Run through each section in the data.
    for section in params:

        # Run through each parameter in this section.
        for option in params[section]:

            if params[section][option] != "":

                try:

                    # Check for continuity between data in file and that in
                    # Longbow.
                    if params[section][option] != oldparams[section][option]:

                        try:

                            # If parameter is changed try adding it to the diff
                            valuediff[section][option] = \
                                params[section][option]

                        except KeyError:

                            # If this is the first time then section won't
                            # exist.
                            valuediff[section] = \
                                {option: params[section][option]}

                # If we get a key error then the paramater is a new one.
                except KeyError:

                    try:

                        # Try adding to diff
                        keydiff[section][option] = params[section][option]

                    except KeyError:

                        # If this is the first time we will need to create the
                        # section.
                        keydiff[section] = {option: params[section][option]}

    # Update the file metastructure with these changes.
    # Firstly handle the updates.
    for section in valuediff:

        # Find the section start (so we know where to look)
        sectionstartindex = contents.index("[" + section + "]")

        # Find the section end.
        try:

            sectionendindex = contents.index(
                [a for a in contents if "[" and "]" in a and
                 contents.index(a) > sectionstartindex][0]) - 1

        except IndexError:

            sectionendindex = len(contents)

        # Limit our search to this range for the parameter.
        for option in valuediff[section]:

            # Get the line index to edit.
            editposition = (
                sectionstartindex +
                contents[sectionstartindex:sectionendindex].index(
                    option + " = " + oldparams[section][option]))

            # Edit the entry.
            contents[editposition] = (
                option + " = " + valuediff[section][option])

    # Now handle new entries. Run through each section.
    for section in keydiff:

        try:
            # Find the section start.
            sectionstartindex = contents.index("[" + section + "]")

            # Find the section end.
            try:
                sectionendindex = contents.index(
                    [a for a in contents if "[" and "]" in a and
                     contents.index(a) > sectionstartindex][0]) - 1

            except IndexError:

                sectionendindex = len(contents)

            # Now for each option.
            for option in keydiff[section]:

                # Insert into the list in the appropriate place.
                contents.insert(
                    sectionendindex, option + " = " + keydiff[section][option])

        # Doesn't exist so it is a new section.
        except ValueError:

            # Append the section.
            contents.extend(["", "[" + section + "]"])

            # Now run through each option.
            for option in keydiff[section]:

                # And append it to the end of the list.
                contents.append(option + " = " + keydiff[section][option])

    try:

        # Open file for writing.
        tmp = open(configfile, "w")

        # Write it all out.
        for item in contents:
            tmp.write(item + "\n")

        # Close file.
        tmp.close()

    except IOError:

        raise EX.ConfigurationError(
            "Error saving to '{0}'".format(configfile))



# Longbow is Copyright (C) of James T Gebbie-Rayet and Gareth B Shannon 2015.
#
# This file is part of the Longbow software which was developed as part of
# the HECBioSim project (http://www.hecbiosim.ac.uk/).
#
# HECBioSim facilitates and supports high-end computing within the
# UK biomolecular simulation community on resources such as ARCHER.
#
# Longbow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Longbow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Longbow.  If not, see <http://www.gnu.org/licenses/>.

"""."""

import logging                                                  # IMPORTANT
import math
import os

try:

    EX = __import__("corelibs.exceptions", fromlist=[''])
    SHELLWRAPPERS = __import__("corelibs.shellwrappers", fromlist=[''])

except ImportError:

    EX = __import__("Longbow.corelibs.exceptions", fromlist=[''])
    SHELLWRAPPERS = __import__("Longbow.corelibs.shellwrappers", fromlist=[''])


# -----------------------------------------------------------------------------
# Boiler plate for logging.

LOGGER = logging.getLogger("Longbow")                           # IMPORTANT

# -----------------------------------------------------------------------------
# For schedulers a unique string to identify the scheduler should go here.
# Sometimes "env | grep -i 'something specific'" will suffice.

QUERY_STRING = "which sbatch"                        # IMPORTANT

# -----------------------------------------------------------------------------
# Delete method (not currently used).


def delete(host, job):                                              # IMPORTANT

    """Method for deleting job."""

    jobid = job["jobid"]

    LOGGER.info("Deleting the job with id '{0}'" .format(jobid))
    try:

        shellout = SHELLWRAPPERS.sendtossh(host, ["scancel " + jobid])

    except EX.SSHError:

        raise EX.JobdeleteError("Unable to delete job.")

    LOGGER.info("Deleted successfully")

    return shellout[0]


# -----------------------------------------------------------------------------
# Prepare method

def prepare(hosts, jobname, jobs):                             # IMPORTANT

    """Create the SLURM jobfile ready for submitting jobs"""

    LOGGER.info("Creating submit file for job '{0}'" .format(jobname))

    # Open file for SLURM script.
    slurmfile = os.path.join(jobs[jobname]["localworkdir"], "submit.slurm")
    jobfile = open(slurmfile, "w+")

    # Write the SLURM script
    jobfile.write("#!/bin/bash --login\n")

    # Job name (if supplied)
    if jobname is not "":

        jobfile.write("#SBATCH -J " + jobname + "\n")

    # Queue to submit to (if supplied)
    if jobs[jobname]["queue"] is not "":

        jobfile.write("#SBATCH -p " + jobs[jobname]["queue"] + "\n")

    resource = jobs[jobname]["resource"]

    # Account to charge (if supplied)
    if hosts[resource]["account"] is not "":

        # if no accountflag is provided use the default
        if hosts[resource]["accountflag"] is "":

            jobfile.write("#SBATCH -A " + hosts[resource]["account"] + "\n")

        else:

            jobfile.write("#SBATCH " +
                          hosts[resource]["accountflag"] +
                          " " + hosts[resource]["account"] + "\n")

    cores = jobs[jobname]["cores"]
    cpn = hosts[resource]["corespernode"]

    # Specify the total number of mpi tasks required
    jobfile.write("#SBATCH -n " + cores + "\n")

    # If user has specified corespernode for under utilisation then
    # set the total nodes (-N) parameter.
    if cpn is not "":

        nodes = float(cores) / float(cpn)

        # Make sure nodes is rounded up to the next highest integer
        nodes = str(int(math.ceil(nodes)))
        jobfile.write("#SBATCH -N " + nodes + "\n")

    # Walltime for job
    jobfile.write("#SBATCH -t " + jobs[jobname]["maxtime"] + ":00\n\n")

    # Load up modules if required.
    if jobs[jobname]["modules"] is not "":

        for module in jobs[jobname]["modules"].split(","):

            module = module.replace(" ", "")
            jobfile.write("module load {0}\n\n" .format(module))

    # Handler that is used for job submission.
    mpirun = hosts[resource]["handler"]

    # Single jobs only need one run command.
    if int(jobs[jobname]["replicates"]) == 1:

        jobfile.write(mpirun + " " + jobs[jobname]["executableargs"] + "\n")

    # Ensemble jobs need a loop.
    elif int(jobs[jobname]["replicates"]) > 1:

        jobfile.write("basedir = `pwd`"
                      "for i in {1.." + jobs[jobname]["replicates"] + "};\n"
                      "do\n"
                      "  cd $basedir/rep$i/\n"
                      "  " + mpirun + " " + jobs[jobname]["executableargs"] +
                      "\n"
                      "done\n"
                      "wait\n")

    # Close the file
    jobfile.close()

    # Append submitfile to list of files ready for staging.
    jobs[jobname]["upload-include"] = (
        jobs[jobname]["upload-include"] + ", submit.slurm")
    jobs[jobname]["subfile"] = "submit.slurm"                     # IMPORTANT


# -----------------------------------------------------------------------------
# Status method

def status(host, jobid):                                        # IMPORTANT

    """Method for querying job."""

    state = ""

    try:

        shellout = SHELLWRAPPERS.sendtossh(host, ["squeue -u " + host["user"] +
                                                  "| grep " + jobid])

        stat = shellout[0].split()

        if stat[4] == "CA":

            state = "Cancelled"

        elif stat[4] == "CD":

            state = "Completed"

        elif stat[4] == "CF":

            state = "Configuring"

        elif stat[4] == "CG":

            state = "Completing"

        elif stat[4] == "F":

            state = "Failed"

        elif stat[4] == "NF":

            state = "Node failure"

        elif stat[4] == "PD":

            state = "Pending"

        elif stat[4] == "PR":

            state = "Preempted"

        elif stat[4] == "R":

            state = "Running"

        elif stat[4] == "S":

            state = "Suspended"

        elif stat[4] == "TO":

            state = "Timed out"

    except EX.SSHError:

        state = "Finished"

    return state

# -----------------------------------------------------------------------------
# Submit method


def submit(host, jobname, jobs):                                # IMPORTANT

    """Method for submitting job."""

    # Set the path to remoteworkdir/jobnamexxxxx
    path = jobs[jobname]["destdir"]

    # Change into the working directory and submit the job.
    cmd = ["cd " + path + "\n" + "sbatch " + jobs[jobname]["subfile"] +
           "| tail -1 | awk '{print $4}'"]

    # Process the submit
    try:

        shellout = SHELLWRAPPERS.sendtossh(host, cmd)[0]

    except EX.SSHError:

        raise EX.JobsubmitError("  Something went wrong when submitting.")

    output = shellout.rstrip("\r\n")

    LOGGER.info("Job '{0}' submitted with id '{1}'" .format(jobname, output))

    jobs[jobname]["jobid"] = output                             # IMPORTANT



# Longbow is Copyright (C) of James T Gebbie-Rayet and Gareth B Shannon 2015.
#
# This file is part of the Longbow software which was developed as part of
# the HECBioSim project (http://www.hecbiosim.ac.uk/).
#
# HECBioSim facilitates and supports high-end computing within the
# UK biomolecular simulation community on resources such as ARCHER.
#
# Longbow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Longbow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Longbow.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains methods for interacting with the Unix shell, it includes
methods for file manipulation and directory functions. Where possible paths
are checked to make sure they are absolute paths.

The following methods can be found:

testconnections()
    This method will test that connections to hosts specified in jobs can be
    established. Problems encountered at this stage could be due to either
    badly configured hosts, networking problems, or even system maintenance/
    downtime on the HPC host.

sendtoshell()
    This method is responsible for handing off commands to the Unix shell, it
    makes use of the subprocess library from the Python standard library.

sendtossh()
    This method constructs a string containing commands to be executed via SSH.
    This string is then handed off to the sendtoshell() method for execution.

sendtorsync()
    This method constructs a string that forms an rsync command, this string is
    then handed off to the sendtoshell() method for execution.

localcopy()
    This method is for copying a file/directory between two local paths, this
    method relies on the Python standard library to perform operations.

localdelete()
    This method is for deleting a file/directory from the local machine, this
    method relies on the Python standard library to perform operations.

locallist()
    This method is for constructing a list of items present within a given
    directory. This method relies on the Python standard library to perform
    operations.

remotecopy()
    This method is for copying a file/directory between two paths on a remote
    host, this is done via passing a copy command to the sendtossh() method.

remotedelete()
    This method is for deleting a file/directory from a path on a remote host,
    this is done via passing a delete command to the sendtossh() method.

remotelist()
    This method is for listing the contents of a directory on a remote host,
    this is done via passing a list command to the sendtoshell() method.

upload()
    This method is for uploading files to a remote host, this method is
    responsible for specifying the direction that the transfer takes place.

download()
    This method is for downloading files from a remote host, this method is
    responsible for specifying the direction that the transfer takes place.
"""

import os
import shutil
import subprocess
import logging
import time

try:

    EX = __import__("corelibs.exceptions", fromlist=[''])

except ImportError:

    EX = __import__("Longbow.corelibs.exceptions", fromlist=[''])

LOGGER = logging.getLogger("Longbow")


def testconnections(hosts, jobs):

    """
    This method will test that connections to hosts specified in jobs can be
    established. Problems encountered at this stage could be due to either
    badly configured hosts, networking problems, or even system maintenance/
    downtime on the HPC host.

    Required arguments are:

    hosts (dictionary) - The Longbow hosts data structure, see configuration.py
                         for more information about the format of this
                         structure.

    jobs (dictionary) - The Longbow jobs data structure, see configuration.py
                        for more information about the format of this
                        structure.
    """

    LOGGER.info(
        "Testing connections to all resources that are referenced in the job "
        "configurations.")

    checked = []

    # Test all of the computers listed in jobs in the job configuration
    # file, there is no need to check all the ones listed in host
    # configuration each time if they are not used.
    for param in jobs:

        resource = jobs[param]["resource"]
        host = hosts[resource]

        # Have we checked this connection already?
        if resource not in checked:

            # Make sure we don't check this again.
            checked.extend([resource])

            LOGGER.debug("Testing connection to '{0}'".format(resource))

            try:

                sendtossh(host, ["ls"])

            except EX.SSHError:

                raise

            LOGGER.info("Test connection to '{0}' - passed".format(resource))


def sendtoshell(cmd):

    """
    This method is responsible for handing off commands to the Unix shell, it
    makes use of the subprocess library from the Python standard library.

    REquired arguments are:

    cmd (string) - A fully qualified Unix command.

    Return parameters are:

    stdout (string) - Contains the output from the standard output of the Unix
                      shell.

    stderr (string) - Contains the output from the standard error of the Unix
                      shell.

    errorstate (string) - Contains the exit code that the Unix shell exits
                          with.
    """

    LOGGER.debug("Sending the following to subprocess '{0}'".format(cmd))

    handle = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)

    stdout, stderr = handle.communicate()

    # Format the string to utf-8 for python 3, python 2 should be untouched.
    stdout = stdout.decode("utf-8")

    # Grab the return code.
    errorstate = handle.returncode

    return stdout, stderr, errorstate


def sendtossh(host, args):

    """
    This method constructs a string containing commands to be executed via SSH.
    This string is then handed off to the sendtoshell() method for execution.

    Required arguments are:

    host (dictionary) - A dictionary containing a single host, this is not to
                        be confused with hosts. The best way to get a single
                        dictionary for a host from hosts is to do:
                        host = hosts[hostname]

    args (list) - A list containing commands to be sent to SSH, multiple
                  commands should each be an entry in the list.

    Return parameters are:

    shellout (tuple of strings) - Contains the three strings returned from the
                                  sendtoshell() method. These are standard
                                  output, standard error and the exit code.
    """

    # basic ssh command.
    cmd = ["ssh", "-p " + host["port"], host["user"] + "@" + host["host"]]

    # add the commands to be sent to ssh.
    cmd.extend(args)

    i = 0

    # This loop is essentially so we can do 3 retries on commands that fail,
    # this is to catch when things go wrong over SSH like dropped connections,
    # issues with latency etc.
    while i is not 3:

        # Send to ssh.
        shellout = sendtoshell(cmd)

        errorstate = shellout[2]

        # If no error exit loop, if errorcode is not 0 raise exception unless
        # code is 255
        if errorstate is 0:

            break

        elif errorstate is 255:

            i = i + 1

        else:

            raise EX.SSHError(
                "SSH failed, make sure a normal terminal can connect to SSH "
                "to be sure there are no connection issues.", shellout)

        # If number of retries hits 3 then give up.
        if i is 3:

            raise EX.SSHError(
                "SSH failed, make sure a normal terminal can connect to SSH "
                "to be sure there are no connection issues.", shellout)

        LOGGER.debug("Retry SSH after 10 second wait.")

        # Wait 10 seconds to see if problem goes away before trying again.
        time.sleep(10)

    return shellout


def sendtorsync(src, dst, port, includemask, excludemask):

    """
    This method constructs a string that forms an rsync command, this string is
    then handed off to the sendtoshell() method for execution.

    Required arguments are:

    src (string) - A string containing the source directory for transfer, if
                   this is a download then this should include the host
                   information. See the download and upload methods for how
                   this should be done (or just make use of those two methods).

    dst (string) - A string containing the destination directory for transfer,
                   if this is an upload then this should include the host
                   information. See the download and upload methods for how
                   this should be done (or just make use of those two methods).

    port (port) - A string containing the port number as to which should be
                  used for transfer.

    includemask (string) - This is a string that should contain a comma
                           separated list of files for transfer.

    excludemask (string) - This is a string that should specify which files
                           should be excluded from rsync transfer, this is
                           useful for not transfering large unwanted files.
    """

    include = []
    exclude = []

    # Figure out if we are using masks to specify files.
    if excludemask is not "" and includemask is "":

        # Exclude masks are a comma separated list.
        for mask in excludemask.split(","):

            mask = mask.replace(" ", "")
            exclude.append("--exclude")
            exclude.append(mask)

        cmd = ["rsync", "-azP"]
        cmd.extend(exclude)
        cmd.extend(["-e", "ssh -p " + port, src, dst])

    elif excludemask is not "" and includemask is not "":

        # Exclude masks are a comma separated list.
        for mask in excludemask.split(","):

            mask = mask.replace(" ", "")
            exclude.append("--exclude")
            exclude.append(mask)

        # Exclude masks are a comma separated list.
        for mask in includemask.split(","):

            mask = mask.replace(" ", "")
            include.append("--include")
            include.append(mask)

        cmd = ["rsync", "-azP"]
        cmd.extend(include)
        cmd.extend(exclude)
        cmd.extend(["-e", "ssh -p " + port, src, dst])

    else:

        # Just normal rsync
        cmd = ["rsync", "-azP", "-e", "ssh -p " + port, src, dst]

    i = 0

    # This loop is essentially so we can do 3 retries on commands that fail,
    # this is to catch when things go wrong over SSH like dropped connections,
    # issues with latency etc.
    while i is not 3:

        # Send to SSH.
        shellout = sendtoshell(cmd)

        errorstate = shellout[2]

        # If no error exit loop, if errorcode is not 0 raise exception unless
        # code is 255
        if errorstate is 0:

            break

        else:

            i = i + 1

        # If number of retries hits 3 then give up.
        if i is 3:

            raise EX.RsyncError(
                "rsync failed, make sure a normal terminal can connect to "
                "rsync to be sure there are no connection issues.", shellout)

        LOGGER.debug("Retry rsync after 10 second wait.")

        # Wait 10 seconds to see if problem goes away before trying again.
        time.sleep(10)


def localcopy(src, dst):

    """
    This method is for copying a file/directory between two local paths, this
    method relies on the Python standard library to perform operations.

    Required arguments are:

    src (string) - A string containing the absolute path of the file/directory
                   to be copied.

    dst (string) - A string containing the destination absolute path to be
                   copied to.
    """

    LOGGER.debug("Copying '{0}' to '{1}'".format(src, dst))

    # Expand tildas (if present) otherwise these will not change anything.
    src = os.path.expanduser(src)
    dst = os.path.expanduser(dst)

    # Are paths absolute.
    if os.path.isabs(src) is False:

        raise EX.AbsolutepathError("The source path is not absolute", src)

    if os.path.isabs(dst) is False:

        raise EX.AbsolutepathError("The destination path is not absolute", dst)

    # Is the source a file or a directory? They are dealt with slightly
    # differently.
    if os.path.isfile(src):

        try:

            # Check if the destination exists.
            if os.path.exists(dst):

                # Copy it.
                shutil.copy(src, dst)

            else:
                os.makedirs(dst)
                shutil.copy(src, dst)

        except (shutil.Error, IOError):

            raise EX.LocalcopyError("Could not copy the file", src)

    elif os.path.isdir(src):

        try:

            # Check if the destination exists.
            if os.path.exists(dst):

                # Remove the existing and then copy it.
                shutil.rmtree(dst)

                shutil.copytree(src, dst)

            else:

                # Copy it.
                shutil.copytree(src, dst)

        except (shutil.Error, IOError):

            raise EX.LocalcopyError("Could not copy the directory '{0}'"
                                    .format(src))


def localdelete(src):

    """
    This method is for deleting a file/directory from the local machine, this
    method relies on the Python standard library to perform operations.

    Required arguments are:

    src (string) - A string containing the absolute path of the file/directory
                   to be deleted.
    """

    LOGGER.debug("Deleting '{0}'".format(src))

    # Expand tildas (if present) otherwise these will not change anything.
    src = os.path.expanduser(src)

    # Check if path is absolute.
    if os.path.isabs(src) is False:

        raise EX.AbsolutepathError("The source path is not absolute", src)

    # Check if we are deleting a file or directory and call the appropriate
    # method.
    if os.path.isfile(src):

        try:

            os.remove(src)

        except IOError:

            raise EX.LocaldeleteError("Could not delete file", src)

    elif os.path.isdir(src):

        try:

            shutil.rmtree(src)

        except IOError:

            raise EX.LocaldeleteError("Could not delete file", src)


def locallist(src):

    """
    This method is for constructing a list of items present within a given
    directory. This method relies on the Python standard library to perform
    operations.

    Required arguments are:

    src (string) - A string containing the absolute path to a directory to
                   be listed.

    Return parameters are:

    filelist (list) - A list of files within the specified directory.
    """

    LOGGER.debug("Listing the contents of '{0}'".format(src))

    # Expand tildas (if present) otherwise these will not change anything.
    src = os.path.expanduser(src)

    # Check if path is absolute.
    if os.path.isabs(src) is False:

        raise EX.AbsolutepathError("The source path is not absolute", src)

    # Check if the path exists, and list if it does.
    if os.path.exists(src):

        filelist = os.listdir(src)

    else:

        raise EX.LocallistError("Local directory does not exist.", src)

    return filelist


def remotecopy(host, src, dst):

    """
    This method is for copying a file/directory between two paths on a remote
    host, this is done via passing a copy command to the sendtossh() method.

    Required arguments are:

    host (dictionary) - A dictionary containing a single host, this is not to
                        be confused with hosts. The best way to get a single
                        dictionary for a host from hosts is to do:
                        host = hosts[hostname]

    src (string) - A string containing the absolute path of the file/directory
                   to be copied (on the host).

    dst (string) - A string containing the destination absolute path to be
                   copied to (on the host).
    """

    LOGGER.debug("Copying '{0}' to '{1}'".format(src, dst))

    # Are paths absolute. Do we start with tildas, if so since we are going
    # through the shell allow it to expand the tilda on the remote host for us.
    if os.path.isabs(src) is False and src[0] != "~":

        raise EX.AbsolutepathError("The source path is not absolute", src)

    if os.path.isabs(dst) is False and dst[0] != "~":

        raise EX.AbsolutepathError("The destination path is not absolute", dst)

    # Just use cp for this with recursive set in case of directory.
    cmd = ["cp -r", src, dst]

    # Send to subprocess.
    try:

        sendtossh(host, cmd)

    except EX.SSHError:

        raise EX.RemotecopyError("Could not copy file to host ", src, dst)


def remotedelete(host, src):

    """
    This method is for deleting a file/directory from a path on a remote host,
    this is done via passing a delete command to the sendtossh() method.

    Required arguments are:

    host (dictionary) - A dictionary containing a single host, this is not to
                        be confused with hosts. The best way to get a single
                        dictionary for a host from hosts is to do:
                        host = hosts[hostname]

    src (string) - A string containing the absolute path of the file/directory
                   to be deleted (on the host).
    """

    LOGGER.debug("Deleting '{0}'".format(src))

    # Are paths absolute.
    if os.path.isabs(src) is False and src[0] != "~":

        raise EX.AbsolutepathError("The source path is not absolute", src)

    # Just use rm for this with recursive set in case of directory.
    cmd = ["rm -r", src]

    # Send to subprocess.
    try:

        sendtossh(host, cmd)

    except EX.SSHError:

        raise EX.RemotedeleteError(
            "Could not delete the file/directory on remote host", src)


def remotelist(host, src):

    """
    This method is for listing the contents of a directory on a remote host,
    this is done via passing a list command to the sendtoshell() method.

    Required arguments are:

    host (dictionary) - A dictionary containing a single host, this is not to
                        be confused with hosts. The best way to get a single
                        dictionary for a host from hosts is to do:
                        host = hosts[hostname]

    src (string) - A string containing the absolute path of the file/directory
                   to be listed (on the host).

    Returned parameters are:

    filelist (list) - A list of files within the specified directory.
    """

    LOGGER.debug("Listing the contents of '{0}'".format(src))

    # Are paths absolute.
    if os.path.isabs(src) is False and src[0] != "~":

        raise EX.AbsolutepathError("The source path is not absolute", src)

    # Shell command ls for listing in a shell.
    cmd = ["ls" + " " + src]

    # Send command to subprocess.
    try:

        shellout = sendtossh(host, cmd)

    except EX.SSHError:

        raise EX.RemotelistError("Could not list the directory", src)

    # Split the stdout into a list.
    filelist = shellout[0].split()

    return filelist


def upload(host, src, dst, includemask, excludemask):

    """
    This method is for uploading files to a remote host, this method is
    responsible for specifying the direction that the transfer takes place.

    Required arguments are:

    host (dictionary) - A dictionary containing a single host, this is not to
                        be confused with hosts. The best way to get a single
                        dictionary for a host from hosts is to do:
                        host = hosts[hostname]

    src (string) - A string containing the source directory for transfer.

    dst (string) - A string containing the destination directory for transfer.

    includemask (string) - This is a string that should contain a comma
                           separated list of files for transfer.

    excludemask (string) - This is a string that should specify which files
                           should be excluded from rsync transfer, this is
                           useful for not transfering large unwanted files.
    """

    # Are paths absolute.
    if os.path.isabs(src) is False and src[0] != "~":

        raise EX.AbsolutepathError("The source path is not absolute", src)

    if os.path.isabs(dst) is False and dst[0] != "~":

        raise EX.AbsolutepathError("The destination path is not absolute", dst)

    dst = (host["user"] + "@" + host["host"] + ":" + dst)

    LOGGER.debug("Copying '{0}' to '{1}'".format(src, dst))

    # Send command to subprocess.
    try:

        sendtorsync(src, dst, host["port"], includemask, excludemask)

    except EX.RsyncError:

        raise


def download(host, src, dst, includemask, excludemask):

    """
    This method is for downloading files from a remote host, this method is
    responsible for specifying the direction that the transfer takes place.

    Required arguments are:

    host (dictionary) - A dictionary containing a single host, this is not to
                        be confused with hosts. The best way to get a single
                        dictionary for a host from hosts is to do:
                        host = hosts[hostname]

    src (string) - A string containing the source directory for transfer.

    dst (string) - A string containing the destination directory for transfer.

    includemask (string) - This is a string that should contain a comma
                           separated list of files for transfer.

    excludemask (string) - This is a string that should specify which files
                           should be excluded from rsync transfer, this is
                           useful for not transfering large unwanted files.
    """

    # Are paths absolute.
    if os.path.isabs(src) is False and src[0] != "~":

        raise EX.AbsolutepathError("The source path is not absolute", src)

    if os.path.isabs(dst) is False and dst[0] != "~":

        raise EX.AbsolutepathError("The destination path is not absolute", dst)

    src = (host["user"] + "@" + host["host"] + ":" + src)

    LOGGER.debug("Copying '{0}' to '{1}'".format(src, dst))

    # Send command to subprocess.
    try:

        sendtorsync(src, dst, host["port"], includemask, excludemask)

    except EX.RsyncError:

        raise



# Longbow is Copyright (C) of James T Gebbie-Rayet and Gareth B Shannon 2015.
#
# This file is part of the Longbow software which was developed as part of
# the HECBioSim project (http://www.hecbiosim.ac.uk/).
#
# HECBioSim facilitates and supports high-end computing within the
# UK biomolecular simulation community on resources such as ARCHER.
#
# Longbow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Longbow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Longbow.  If not, see <http://www.gnu.org/licenses/>.

"""

Corelibs - the Longbow library

"""



# Longbow is Copyright (C) of James T Gebbie-Rayet and Gareth B Shannon 2015.
#
# This file is part of the Longbow software which was developed as part of
# the HECBioSim project (http://www.hecbiosim.ac.uk/).
#
# HECBioSim facilitates and supports high-end computing within the
# UK biomolecular simulation community on resources such as ARCHER.
#
# Longbow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Longbow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Longbow.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains methods for setting up and configuring the standard
logging module provided by the python standard library. There are different
types of logger configuration within this module, and these types lend
themselves to fulfil the requirements of Longbow.

The following methods can be found in this module:

setuplogger()
    This method is used by Longbow to determine which level of logging is
    required based on user input. This method is specific to the function
    of Longbow and may or may not be suitable for developers to use in their
    own software.

standardlogger()
    This method configures the python logger to log to the specified file,
    the standard logging profile only writes out messages marked at info,
    warning and error level. Statements in the code that are marked as debug
    will be hidden by this logger.

verboselogger()
    This method configures the python logger to log to the specified file along
    with placing the same information into the standard output (usually a
    terminal or console), the verbose logging profile only writes out messages
    marked at info, warning and error level. Statements in the code that are
    marked as debug will be hidden by this logger.

debuglogger()
    This method configures the python logger to log to the specified file along
    with placing the same information into the standard output (usually a
    terminal or console). The debug logging profile will write out messages
    marked under all levels, this is to enable maximum information for
    debugging purposes.
"""

import logging


def setuplogger(logfile, loggername, mode):

    """
    This method is used by Longbow to determine which level of logging is
    required based on user input. This method is specific to the function
    of Longbow and may or may not be suitable for developers to use in their
    own software.

    Required arguments are:

    logfile (string) - path to file that will be written to.

    loggername (string) - Longbow modules all look for a module called
                          "longbow", so this should be the same if you are
                          using this to configure a longbow logger.

    mode (dictionary) -  A dictionary that contains true/false pairs for the
                         parameters debug and verbose, if all are false then
                         a standard logger will be set up.
    """

    if mode["debug"]:

        debuglogger(logfile, loggername)

    elif mode["verbose"]:

        verboselogger(logfile, loggername)

    else:

        standardlogger(logfile, loggername)


def standardlogger(logfile, loggername):

    """
    This method configures the python logger to log to the specified file,
    the standard logging profile only writes out messages marked at info,
    warning and error level. Statements in the code that are marked as debug
    will be hidden by this logger.

    Required arguments are:

    logfile (string) - path to file that will be written to.

    loggername (string) - Longbow modules all look for a module called
                          "longbow", so this should be the same if you are
                          using this to configure a longbow logger.
    """

    # Create a logger.
    logger = logging.getLogger(loggername)
    logger.setLevel(logging.INFO)

    # Define a logging format.
    logformat = logging.Formatter("%(asctime)s - %(message)s")

    # Set logger to write to file, set level and bind format.
    loghandle = logging.FileHandler(logfile, mode="w")
    loghandle.setLevel(logging.INFO)
    loghandle.setFormatter(logformat)
    logger.addHandler(loghandle)


def verboselogger(logfile, loggername):

    """
    This method configures the python logger to log to the specified file along
    with placing the same information into the standard output (usually a
    terminal or console), the verbose logging profile only writes out messages
    marked at info, warning and error level. Statements in the code that are
    marked as debug will be hidden by this logger.

    Required arguments are:

    logfile (string) - path to file that will be written to.

    loggername (string) - Longbow modules all look for a module called
                          "longbow", so this should be the same if you are
                          using this to configure a longbow logger.
    """

    # Create a logger.
    logger = logging.getLogger(loggername)
    logger.setLevel(logging.INFO)

    # Define a logging format.
    logformat = logging.Formatter("%(asctime)s - %(message)s")

    # Default is to write to the log file, set level and bind format.
    loghandler = logging.FileHandler(logfile, mode="w")
    loghandler.setLevel(logging.INFO)
    loghandler.setFormatter(logformat)
    logger.addHandler(loghandler)

    loghandler = logging.StreamHandler()
    loghandler.setLevel(logging.INFO)
    loghandler.setFormatter(logformat)
    logger.addHandler(loghandler)


def debuglogger(logfile, loggername):

    """
    This method configures the python logger to log to the specified file along
    with placing the same information into the standard output (usually a
    terminal or console). The debug logging profile will write out messages
    marked under all levels, this is to enable maximum information for
    debugging purposes.

    Required arguments are:

    logfile (string) - path to file that will be written to.

    loggername (string) - Longbow modules all look for a module called
                          "longbow", so this should be the same if you are
                          using this to configure a longbow logger.
    """

    # Create a logger.
    logger = logging.getLogger(loggername)
    logger.setLevel(logging.DEBUG)

    # Define a logging format.
    logformat = logging.Formatter("%(asctime)s - %(name)s - " +
                                  "%(filename)-18s - %(levelname)-8s" +
                                  " -   %(message)s")

    # Default is to write to the log file, set level and bind format.
    loghandler = logging.FileHandler(logfile, mode="w")
    loghandler.setLevel(logging.DEBUG)
    loghandler.setFormatter(logformat)
    logger.addHandler(loghandler)

    loghandler = logging.StreamHandler()
    loghandler.setLevel(logging.DEBUG)
    loghandler.setFormatter(logformat)
    logger.addHandler(loghandler)



# Longbow is Copyright (C) of James T Gebbie-Rayet and Gareth B Shannon 2015.
#
# This file is part of the Longbow software which was developed as part of
# the HECBioSim project (http://www.hecbiosim.ac.uk/).
#
# HECBioSim facilitates and supports high-end computing within the
# UK biomolecular simulation community on resources such as ARCHER.
#
# Longbow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Longbow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Longbow.  If not, see <http://www.gnu.org/licenses/>.

"""
This module provides the basic import framwork for Longbow plug-ins. On import,
this module will import an plug-in packages that are contained within the
plug-ins directory. This gives direct access to any new plug-ins that are added
to the directory for use within the core library.
"""

import pkgutil
import os
import sys

PATH = os.path.dirname(__file__)
PACKAGES = pkgutil.walk_packages(path=[PATH])
PLUGINS = {}

for packageloader, packagename, ispkg in PACKAGES:
    try:
        package = __import__("Longbow.plugins." + packagename)

    except ImportError:
        package = __import__("plugins." + packagename)

    PLUGINS[packagename] = "plugins." + packagename



# Longbow is Copyright (C) of James T Gebbie-Rayet and Gareth B Shannon 2015.
#
# This file is part of the Longbow software which was developed as part of
# the HECBioSim project (http://www.hecbiosim.ac.uk/).
#
# HECBioSim facilitates and supports high-end computing within the
# UK biomolecular simulation community on resources such as ARCHER.
#
# Longbow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Longbow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Longbow.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the 'custom' exception classes for the Longbow core
library. These exceptions are best used in methods that replace/override those
of the core library, such that a developer can make use of the error handling
framework more organically.
"""

# -----------------------------------------------------------------------------
# Exceptions for paths


class AbsolutepathError(Exception):

    """
    Exception class for absolute path errors, usage

    AbsolutepathError(message, path)
    """

    def __init__(self, message, path):

        # Call the base class constructor.
        super(AbsolutepathError, self).__init__(message)

        # Bind the standard outputs.
        self.path = path


class DirectorynotfoundError(Exception):

    """
    Directory not found exception, to provide informative exceptions
    of why a method has failed when directory cannot be found.
    """

    pass

# -----------------------------------------------------------------------------
# Exceptions for applications.py


class CommandlineargsError(Exception):

    """
    Command-line arguments error.
    """

    def __init__(self, message):

        # Call the base class constructor.
        super(CommandlineargsError, self).__init__(message)


class RequiredinputError(Exception):

    """
    Required input error exception.
    """

    def __init__(self, message):

        # Call the base class constructor.
        super(RequiredinputError, self).__init__(message)

# -----------------------------------------------------------------------------
# Exceptions for configuration.py


class ConfigurationError(Exception):

    """
    Configuration error.
    """

    pass

# -----------------------------------------------------------------------------
# Exceptions for plugin.py


class PluginattributeError(Exception):

    """
    Missing plugin method exception.
    """

    pass


# -----------------------------------------------------------------------------
# Exceptions for scheduling.py


class SchedulercheckError(Exception):

    """
    Scheduler checking exception.
    """

    pass


class HandlercheckError(Exception):

    """
    Job handler checking exception.
    """

    pass


class JobdeleteError(Exception):

    """
    Job delete exception.
    """

    pass


class JobsubmitError(Exception):

    """
    Job submit exception.
    """

    pass

# -----------------------------------------------------------------------------
# Exceptions for shellwrappers.py


class LocalcopyError(Exception):

    """
    Copy on local machine exception.
    """

    def __init__(self, message, path):

        # Call the base class constructor.
        super(LocalcopyError, self).__init__(message)

        # Bind the standard outputs.
        self.path = path


class LocaldeleteError(Exception):

    """
    Delete on local machine exception.
    """

    def __init__(self, message, path):

        # Call the base class constructor.
        super(LocaldeleteError, self).__init__(message)

        # Bind the standard outputs.
        self.path = path


class LocallistError(Exception):

    """
    List on local machine exception.
    """

    def __init__(self, message, path):

        # Call the base class constructor.
        super(LocallistError, self).__init__(message)

        # Bind the standard outputs.
        self.path = path


class ProtocolError(Exception):

    """
    Unknown protocol exception.
    """

    def __init__(self, message, protocol):

        # Call the base class constructor.
        super(ProtocolError, self).__init__(message)

        # Bind the standard outputs.
        self.path = protocol


class RemotecopyError(Exception):

    """
    Copy on remote machine exception.
    """

    def __init__(self, message, src, dst):

        # Call the base class constructor.
        super(RemotecopyError, self).__init__(message)

        # Bind the standard outputs.
        self.src = src
        self.dst = dst


class RemotedeleteError(Exception):

    """
    Delete on remote machine exception.
    """

    def __init__(self, message, path):

        # Call the base class constructor.
        super(RemotedeleteError, self).__init__(message)

        # Bind the standard outputs.
        self.path = path


class RemotelistError(Exception):

    """
    List on remote machine exception.
    """

    def __init__(self, message, path):

        # Call the base class constructor.
        super(RemotelistError, self).__init__(message)

        # Bind the standard outputs.
        self.path = path


class RsyncError(Exception):

    """
    Rsync exception.
    """

    def __init__(self, message, shellout):

        # Call the base class constructor.
        super(RsyncError, self).__init__(message)

        # Bind the standard outputs.
        self.errorcode = shellout[2]
        self.stdout = shellout[0]
        self.stderr = shellout[1]


class SCPError(Exception):

    """
    SCP exception
    """

    def __init__(self, message, shellout):

        # Call the base class constructor.
        super(SCPError, self).__init__(message)

        # Bind the standard outputs.
        self.errorcode = shellout[2]
        self.stdout = shellout[0]
        self.stderr = shellout[1]


class SSHError(Exception):

    """
    SSH exception
    """

    def __init__(self, message, shellout):

        # Call the base class constructor.
        super(SSHError, self).__init__(message)

        # Bind the standard outputs.
        self.errorcode = shellout[2]
        self.stdout = shellout[0]
        self.stderr = shellout[1]

# -----------------------------------------------------------------------------
# Exceptions for staging.py


class StagingError(Exception):

    """
    Generic staging error exception
    """

    pass


class RemoteworkdirError(Exception):

    """
    Exception issued when destdir is the same path as remoteworkdir
    """

    pass



#!/usr/bin/env python

# Longbow is Copyright (C) of James T Gebbie-Rayet and Gareth B Shannon 2015.
#
# This file is part of the Longbow software which was developed as part of
# the HECBioSim project (http://www.hecbiosim.ac.uk/).
#
# HECBioSim facilitates and supports high-end computing within the
# UK biomolecular simulation community on resources such as ARCHER.
#
# Longbow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Longbow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Longbow.  If not, see <http://www.gnu.org/licenses/>.

"""Setup script. Used by easy_install and pip."""

from distutils.core import setup
import os
import sys

# Check for unsupported Python versions.
major = sys.version_info[0]
minor = sys.version_info[1]

if not (major >= 2 and minor >= 6):

    print('The Python version installed is "{0}.{1}", Longbow does not support '
          'this version. We recommend that you install at least version 2.7'
          .format(major, minor))

else:

    print('The Python version installed "{0}.{1}" is supported by Longbow.'
          .format(major, minor))

try:
    # Setting up the .Longbow directory.
    if not os.path.isdir(os.path.expanduser('~/.Longbow')):

        print('Longbow will create a hidden directory in your $HOME directory '
              'in which it will create the hosts configuration file. You will '
              'need to edit this file with your account information on the '
              'HPC machines you wish to use. See documentation for more '
              'information - www.hecbiosim.ac.uk/longbow-docs')

        os.mkdir(os.path.expanduser('~/.Longbow'))
    
        hostfile = open(os.path.expanduser('~/.Longbow/hosts.conf', 'w+'))
    
        hostfile.write('[QuickStart]')
        hostfile.write('host = login.hpc.ac.uk')
        hostfile.write('user = myusername')
        hostfile.write('corespernode = 24')
        hostfile.write('cores = 24')
        hostfile.write('remoteworkdir = /work/myusername/')
        hostfile.write('account = myaccount')
        hostfile.write('modules = mymodules')

        hostfile.close()

except:
    
    print('Longbow failed to create the host configuration file in '
          '"~/.Longbow/hosts.conf", you will have to do this manually. The '
          'user documentation details the information that should be in this '
          'file.')       

# Setup
setup(name = 'Longbow',
      version = '1.1.6',
      description = 'Biomolecular simulation remote job submission tool.',
      long_description = open('README.rst').read(),
      author = 'James T Gebbie-Rayet, Gareth B Shannon',
      author_email = ('james.gebbie@stfc.ac.uk, '
                      'gareth.shannon@nottingham.ac.uk'),
      url = 'www.hecbiosim.ac.uk',
      license = 'OSI Approved :: GNU General Public License v2 (GPLv2)',
      classifiers = ['Development Status :: 5 - Production/Stable',
                     'Environment :: Console',
                     'Programming Language :: Python',
                     'Programming Language :: Python :: 2.6',
                     'Programming Language :: Python :: 2.7',
                     'Programming Language :: Python :: 3',
                     'Topic :: Scientific/Engineering',
                     'Topic :: Utilities',
                     'Operating System :: MacOS :: MacOS X',
                     'Operating System :: POSIX',
                     'Operating System :: Unix'
                     ],
      packages = ['Longbow', 'Longbow.plugins', 'Longbow.plugins.schedulers',
                  'Longbow.plugins.apps', 'Longbow.corelibs'],
      scripts = ['Longbow/longbow'],
      )



# Longbow is Copyright (C) of James T Gebbie-Rayet and Gareth B Shannon 2015.
#
# This file is part of the Longbow software which was developed as part of
# the HECBioSim project (http://www.hecbiosim.ac.uk/).
#
# HECBioSim facilitates and supports high-end computing within the
# UK biomolecular simulation community on resources such as ARCHER.
#
# Longbow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Longbow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Longbow.  If not, see <http://www.gnu.org/licenses/>.

"""."""

import logging
import os
import math

try:

    EX = __import__("corelibs.exceptions", fromlist=[''])
    SHELLWRAPPERS = __import__("corelibs.shellwrappers", fromlist=[''])

except ImportError:

    EX = __import__("Longbow.corelibs.exceptions", fromlist=[''])
    SHELLWRAPPERS = __import__("Longbow.corelibs.shellwrappers", fromlist=[''])

LOGGER = logging.getLogger("Longbow")

QUERY_STRING = "env | grep -i 'pbs'"


def delete(host, job):

    """Method for deleting job."""

    jobid = job["jobid"]

    LOGGER.info("Deleting the job with id '{0}'" .format(jobid))

    try:

        if int(job["replicates"]) > 1:

            shellout = SHELLWRAPPERS.sendtossh(host, ["qdel " + jobid + "[]"])

        else:

            shellout = SHELLWRAPPERS.sendtossh(host, ["qdel " + jobid])

    except EX.SSHError:

        raise EX.JobdeleteError("Unable to delete job.")

    LOGGER.info("Deletion successful.")

    return shellout[0]


def prepare(hosts, jobname, jobs):

    """Create the PBS jobfile ready for submitting jobs"""

    LOGGER.info("Creating submit file for job '{0}'" .format(jobname))

    # Open file for PBS script.
    pbsfile = os.path.join(jobs[jobname]["localworkdir"], "submit.pbs")
    jobfile = open(pbsfile, "w+")

    # Write the PBS script
    jobfile.write("#!/bin/bash --login\n")

    # Job name (if supplied)
    if jobname is not "":

        jobfile.write("#PBS -N " + jobname + "\n")

    # Queue to submit to (if supplied)
    if jobs[jobname]["queue"] is not "":

        jobfile.write("#PBS -q " + jobs[jobname]["queue"] + "\n")

    # Account to charge (if supplied).
    if hosts[jobs[jobname]["resource"]]["account"] is not "":

        # if no accountflag is provided use the default
        if hosts[jobs[jobname]["resource"]]["accountflag"] is "":

            jobfile.write("#PBS -A " +
                          hosts[jobs[jobname]["resource"]]["account"] +
                          "\n")

        # else use the accountflag provided
        else:

            jobfile.write("#PBS " +
                          hosts[jobs[jobname]["resource"]]["accountflag"] +
                          " " + hosts[jobs[jobname]["resource"]]["account"] +
                          "\n")

    cpn = hosts[jobs[jobname]["resource"]]["corespernode"]

    cores = jobs[jobname]["cores"]

    # Load levelling override. In cases where # of cores is less than
    # corespernode, user is likely to be undersubscribing.
    if int(cores) < int(cpn):

        cpn = cores

    # Calculate the number of nodes.
    nodes = float(cores) / float(cpn)

    # Makes sure nodes is rounded up to the next highest integer.
    nodes = str(int(math.ceil(nodes)))

    # Number of cpus per node (most machines will charge for all available cpus
    # per node whether you are using them or not)
    ncpus = cpn

    # Number of mpi processes per node.
    mpiprocs = cpn

    # Memory size (used to select nodes with minimum memory).
    memory = jobs[jobname]["memory"]

    tmp = "select=" + nodes + ":ncpus=" + ncpus + ":mpiprocs=" + mpiprocs

    # If user has specified memory append the flag (not all machines support
    # this).
    if memory is not "":

        tmp = tmp + ":mem=" + memory + "gb"

    # Write the resource requests
    jobfile.write("#PBS -l " + tmp + "\n")

    # Walltime for job.
    jobfile.write("#PBS -l walltime=" + jobs[jobname]["maxtime"] + ":00\n")

    # Set up replicates jobs
    if int(jobs[jobname]["replicates"]) > 1:

        jobfile.write("#PBS -J 1-" + jobs[jobname]["replicates"] + "\n")
        jobfile.write("#PBS -r y\n")

    # Set some environment variables for PBS.
    jobfile.write("\n"
                  "export PBS_O_WORKDIR=$(readlink -f $PBS_O_WORKDIR)\n"
                  "cd $PBS_O_WORKDIR\n"
                  "export OMP_NUM_THREADS=1\n"
                  "\n")

    # Load up modules if required.
    if jobs[jobname]["modules"] is not "":

        for module in jobs[jobname]["modules"].split(","):

            module = module.replace(" ", "")
            jobfile.write("module load {0}\n\n" .format(module))

    # Handler that is used for job submission.
    mpirun = hosts[jobs[jobname]["resource"]]["handler"]

    # CRAY's use aprun which has slightly different requirements to mpirun.
    if mpirun == "aprun":

        mpirun = mpirun + " -n " + cores + " -N " + mpiprocs

    # Single jobs only need one run command.
    if int(jobs[jobname]["replicates"]) == 1:

        jobfile.write(mpirun + " " + jobs[jobname]["executableargs"] + "\n")

    # Job array
    elif int(jobs[jobname]["replicates"]) > 1:

        jobfile.write("basedir=$PBS_O_WORKDIR \n"
                      "cd $basedir/rep${PBS_ARRAY_INDEX}/\n\n" +
                      mpirun + " " + jobs[jobname]["executableargs"] + "\n")

    # Close the file (housekeeping)
    jobfile.close()

    # Append pbs file to list of files ready for staging.
    jobs[jobname]["upload-include"] = (
        jobs[jobname]["upload-include"] + ", submit.pbs")
    jobs[jobname]["subfile"] = "submit.pbs"


def status(host, jobid):

    """Method for querying job."""

    state = ""

    try:

        shellout = SHELLWRAPPERS.sendtossh(host, ["qstat -u " + host["user"] +
                                                  " | grep " + jobid])

        stat = shellout[0].split()

        if stat[9] == "H":

            state = "Held"

        elif stat[9] == "Q":

            state = "Queued"

        elif stat[9] == "R":

            state = "Running"

        elif stat[9] == "B":

            state = "Subjob(s) running"

        elif stat[9] == "E":

            state = "Exiting"

        elif stat[9] == "M":

            state = "Job moved to server"

        elif stat[9] == "S":

            state = "Suspended"

        elif stat[9] == "T":

            state = "Job moved to new location"

        elif stat[9] == "U":

            state = ("Cycle-harvesting job is suspended due to keyboard " +
                     "activity")

        elif stat[9] == "W":

            state = "Waiting for start time"

        elif stat[9] == "X":

            state = "Subjob completed execution/has been deleted"

    except EX.SSHError:

        state = "Finished"

    return state


def submit(host, jobname, jobs):

    """Method for submitting a job."""

    # Set the path to remoteworkdir/jobnamexxxxx
    path = jobs[jobname]["destdir"]

    # Change into the working directory and submit the job.
    cmd = ["cd " + path + "\n", "qsub " + jobs[jobname]["subfile"] +
           " | grep -P -o '[0-9]*(?=.)'"]

    # Process the submit
    try:

        shellout = SHELLWRAPPERS.sendtossh(host, cmd)[0]

    except EX.SSHError as inst:

        if "set_booleans" in inst.stderr:

            raise EX.JobsubmitError(
                "Something went wrong when submitting. The likely cause is "
                "your particular PBS install is not receiving the "
                "information/options/parameters it " "requires "
                "e.g. '#PBS -l mem=20gb'. Check the PBS documentation and edit"
                " the configuration files to provide the necessary information"
                "e.g. 'memory = 20' in the job configuration file")

        elif "Job rejected by all possible destinations" in inst.stderr:

            raise EX.JobsubmitError(
                "Something went wrong when submitting. This may be because "
                "you need to provide PBS with your account code and the "
                "account flag your PBS install expects (Longbow defaults to "
                "A). Check the PBS documentation and edit the configuration "
                "files to provide the necessary information e.g. "
                "'accountflag = P' and 'account = ABCD-01234-EFG'")

        elif "Job must specify budget (-A option)" in inst.stderr:

            raise EX.JobsubmitError(
                "Something went wrong when submitting. This may be because "
                "you provided PBS with an account flag other than 'A' which "
                "your PBS install expects")

        elif "Job exceeds queue and/or server resource limits" in inst.stderr:

            raise EX.JobsubmitError(
                "Something went wrong when submitting. PBS has reported "
                "that 'Job exceeds queue and/or server resource limits'. "
                "This may be because you set a walltime or some other "
                "quantity that exceeds the maximum allowed on your system.")

        elif "budget" in inst.stderr:

            raise EX.JobsubmitError(
                "Something went wrong when submitting. This may be that you "
                "have entered an incorrect account code.")

        elif "illegal -N value" in inst.stderr:

            raise EX.JobsubmitError(
                "Something went wrong when submitting. This is due to the job "
                "name being too long, consult your system administrators/"
                "documentation to query this policy (try < 15 chars).")

        else:

            raise EX.JobsubmitError("Something went wrong when submitting.")

    output = shellout.rstrip("\r\n")

    LOGGER.info("Job '{0}' submitted with id '{1}'" .format(jobname, output))

    jobs[jobname]["jobid"] = output



# Longbow is Copyright (C) of James T Gebbie-Rayet and Gareth B Shannon 2015.
#
# This file is part of the Longbow software which was developed as part of
# the HECBioSim project (http://www.hecbiosim.ac.uk/).
#
# HECBioSim facilitates and supports high-end computing within the
# UK biomolecular simulation community on resources such as ARCHER.
#
# Longbow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Longbow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Longbow.  If not, see <http://www.gnu.org/licenses/>.

"""."""

import sys
import os
import pkgutil
import logging

QUERY = {}

PATH = os.path.dirname(__file__)
MODULES = pkgutil.iter_modules(path=[PATH])

for loader, modulename, ispkg in MODULES:

    if modulename not in sys.modules:
        try:
            mod = __import__("Longbow.plugins.schedulers." + modulename,
                             fromlist=[""])

        except ImportError:
            try:
                mod = __import__("plugins.schedulers." + modulename,
                                 fromlist=[""])

            except ImportError:
                raise

        try:
            QUERY[modulename] = [getattr(mod, "QUERY_STRING")]

        except AttributeError:
            raise



# Longbow is Copyright (C) of James T Gebbie-Rayet and Gareth B Shannon 2015.
#
# This file is part of the Longbow software which was developed as part of
# the HECBioSim project (http://www.hecbiosim.ac.uk/).
#
# HECBioSim facilitates and supports high-end computing within the
# UK biomolecular simulation community on resources such as ARCHER.
#
# Longbow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Longbow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Longbow.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the import logic for the applications plug-in. It makes
available all the methods. All code for a new plug-in should be placed inside
the plug-in module itself and not here, follow the template for constructing
new app plug-ins.
"""

import logging
import os
import pkgutil
import sys

PATH = os.path.dirname(__file__)
MODULES = pkgutil.iter_modules(path=[PATH])

# These structures make using the code lighter on the utilisation side
# devs can get the data they want using getattr without having to construct
# there own structures. This is at least true for basic lists and dicts.
APPDATA = {}
EXECLIST = []
DEFMODULES = {}
EXECFLAGS = {}

# Optional param for if the modules are named in a different way to simply
# that the modules are called the same as the software (rare).
MODULEOVERIDES = {}
# Loop through all the modules in the plugin.
for loader, modulename, ispkg in MODULES:

    # check for double loading in the namespace.
    if modulename not in sys.modules:
        # try to import using the pip package path.
        try:
            mod = __import__(
                "Longbow.plugins.apps." + modulename, fromlist=[""])

        except ImportError:
            # Else try to import using the non packaged path.
            try:
                mod = __import__(
                    "plugins.apps." + modulename, fromlist=[""])
            except ImportError:
                # Otherwise we've had it! Raise exception.
                raise

        # Now try and pull in attributes.
        try:
            APPDATA[modulename] = getattr(mod, "EXECDATA")

        except AttributeError:
            raise

        try:
            MODULEOVERIDES[modulename] = getattr(mod, "MODULEOVERIDE")

        except AttributeError:
            MODULEOVERIDES[modulename] = ""

# Construct common structures to make programmers lives easier.
for plugin in APPDATA:
    for executable, flags in APPDATA[plugin].items():
        # compile a list of executables.
        EXECLIST.append(executable)

        # compile dictionary of required input flags.
        EXECFLAGS[executable] = flags

        # Compile a list of default modules.
        if MODULEOVERIDES[plugin] == "":
            DEFMODULES[executable] = plugin
        else:
            DEFMODULES[executable] = MODULEOVERIDES[plugin]



# Longbow is Copyright (C) of James T Gebbie-Rayet and Gareth B Shannon 2015.
#
# This file is part of the Longbow software which was developed as part of
# the HECBioSim project (http://www.hecbiosim.ac.uk/).
#
# HECBioSim facilitates and supports high-end computing within the
# UK biomolecular simulation community on resources such as ARCHER.
#
# Longbow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Longbow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Longbow.  If not, see <http://www.gnu.org/licenses/>.

"""."""

import os
import re

try:

    EX = __import__("corelibs.exceptions", fromlist=[''])

except ImportError:

    EX = __import__("Longbow.corelibs.exceptions", fromlist=[''])


EXECDATA = {
    "lmp_xc30": ["-i"]
    }


def file_parser(filename, path, files, substitutions=None):
    '''
    Recursive function that will assimilate from lammps input files a list of
    files (files) to be staged to the execution host. filename will be added
    to the list and any files mentioned in filename will also be added and
    searched. Substitutions is a dictionary of "$" style variables.
    '''

    # Check the location of filename
    addfile = ""

    # if the filename has an absolute path but doesn't exist locally, assume
    # it is on the remote resource
    if os.path.isabs(filename) is True:

        if os.path.isfile(filename) is False:

            addfile = ""

        else:

            raise EX.RequiredinputError(
                "It appears that the user is trying to refer to a file %s "
                "using an explicit path. Please just provide the names of "
                "input files" % filename)

    # elif the file is in the given path
    elif os.path.isfile(os.path.join(path, filename)) is True:

        addfile = filename

    # else issue a warning
    else:

        raise EX.RequiredinputError(
            "It appears the file %s is not present in the expected"
            " directory." % filename)

    # Now look for references to other files in the input file if not done so
    # already
    if addfile and (addfile not in files or not files):

        files.append(addfile)

        # Create a dictionary for any variable substitutions
        # Define keywords and create a dictionary for variable substitutions
        keywords = ['read_data', 'read_restart', 'read_dump']
        variables = {} if not substitutions else substitutions

        fil = None
        # Open the file
        try:

            fil = open(os.path.join(path, addfile), "r")

        except IOError:

            EX.RequiredinputError("Can't read the %s file:" % addfile)

        if fil:

            # search every line for possible input files
            for line in fil:

                # if line commented out, skip
                if line[0] == "#":

                    continue

                # Remove comments
                if '#' in line:

                    end = line.index('#')

                else:

                    end = len(line)

                words = line[:end].split()

                if len(words) > 0:

                    # allow substitutions from inside the input file as well
                    if words[0].lower() == 'variable':

                        variables[words[1]] = words[3]

                    # if this line is reading in an input file
                    if words[0].lower() in keywords:

                        newfile = words[1]

                        # Do variable substitution
                        if '$' in newfile and len(variables.keys()) > 0:

                            start = newfile.index('$')+1

                            if newfile[start] == '{':

                                end = newfile[start:].index('}')+start
                                var = variables[newfile[start+1:end]]
                                newfile = (newfile[0:start-1] + var +
                                           newfile[end+1:])

                            else:

                                end = start+1
                                var = variables[newfile[start:end]]
                                newfile = (newfile[0:start-1] + var +
                                           newfile[end:])

                        # deduce the location of newfile
                        newpath = path

                        if newfile.count("../") == 1:

                            # if we are in a repX subdirectory, the file must
                            # be in cwd
                            if re.search(r'rep\d', addfile):

                                _, _, after = newfile.rpartition("/")
                                newfile = after

                            # else we must be in cwd so issue a warning about
                            # referring to a file that is above cwd
                            else:

                                raise EX.RequiredinputError(
                                    "It appears that the"
                                    " user is trying to refer to a file '{0}'"
                                    " in file '{1}' that is a"
                                    " directory up from the '{2}'"
                                    " directory. Only files in '{3}'"
                                    " or a repX subdirectory can be"
                                    " copied to the remote resource. If the "
                                    "file you are trying to refer to is on the"
                                    " remote resource, give"
                                    " the explicit path to the file.".format(
                                        newfile, addfile, path, path))

                        # elif ../../ is used in an input script issue an error
                        elif newfile.count("../") > 1:

                            raise EX.RequiredinputError(
                                "It appears that the user is trying to refer "
                                "to a file '{0}' in file '{1}' that's multiple"
                                " directories up from a valid directory. This "
                                "is not permitted. If the file you are trying "
                                "to refer to is on the remote resource, give "
                                "the explicit "
                                "path to the file.".format(newfile, addfile))

                        # elif we are in a repX subdirectory and the file
                        # isn't in ../ or ./repX, the file is presumably in the
                        # same directory
                        elif (re.search(r'rep\d', addfile) and not
                              re.search(r'rep\d', newfile)):

                            splitpath, _ = os.path.split(addfile)
                            newfile = os.path.join(splitpath, newfile)

                        # elif newfile is indicated to be in a repX
                        # subdirectory...
                        elif re.search(r'rep\d', newfile):

                            # if we are already in a repX subdirectory issue a
                            # warning
                            if re.search(r'rep\d', addfile):

                                raise EX.RequiredinputError(
                                    "It appears that the"
                                    "user is trying to refer to a file '{0}'"
                                    " that is in a repX/repX "
                                    "subdirectory. This is not "
                                    "permitted.".format(newfile))

                            # else we must be in cwd...
                            else:

                                newfile = "rep" + newfile.split("rep")[1]

                        # recursive function
                        file_parser(newfile, newpath, files, substitutions)

            fil.close()


def sub_dict(args):
    '''
    Function to detect substitutions specified on the commandline.
    '''

    removelist = []
    sub = {}

    for index, item in enumerate(args):

        if item == "-var" or item == "-v":

            sub[args[index + 1]] = args[index + 2]
            removelist.append(item)
            removelist.append(args[index + 1])
            removelist.append(args[index + 2])

    for item in removelist:

        args.remove(item)

    return sub



# Longbow is Copyright (C) of James T Gebbie-Rayet and Gareth B Shannon 2015.
#
# This file is part of the Longbow software which was developed as part of
# the HECBioSim project (http://www.hecbiosim.ac.uk/).
#
# HECBioSim facilitates and supports high-end computing within the
# UK biomolecular simulation community on resources such as ARCHER.
#
# Longbow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Longbow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Longbow.  If not, see <http://www.gnu.org/licenses/>.

"""."""

import logging
import os

try:

    EX = __import__("corelibs.exceptions", fromlist=[''])
    SHELLWRAPPERS = __import__("corelibs.shellwrappers", fromlist=[''])

except ImportError:

    EX = __import__("Longbow.corelibs.exceptions", fromlist=[''])
    SHELLWRAPPERS = __import__("Longbow.corelibs.shellwrappers", fromlist=[''])

LOGGER = logging.getLogger("Longbow")

QUERY_STRING = "env | grep -i 'lsf'"


def delete(host, job):

    """Method for deleting job."""

    jobid = job["jobid"]

    LOGGER.info("Deleting the job with id '{0}'" .format(jobid))

    try:

        shellout = SHELLWRAPPERS.sendtossh(host, ["bkill " + jobid])

    except EX.SSHError:

        raise EX.JobdeleteError("Unable to delete job.")

    LOGGER.info("Deletion successful")

    return shellout[0]


def prepare(hosts, jobname, jobs):

    """Create the LSF jobfile ready for submitting jobs"""

    LOGGER.info("Creating submit file for job '{0}'" .format(jobname))

    # Open file for LSF script.
    lsffile = os.path.join(jobs[jobname]["localworkdir"], "submit.lsf")
    jobfile = open(lsffile, "w+")

    # Write the PBS script
    jobfile.write("#!/bin/bash --login\n")

    if jobname is not "":

        # Single job
        if int(jobs[jobname]["replicates"]) == 1:

            jobfile.write("#BSUB -J " + jobname + "\n")

        # Job array
        elif int(jobs[jobname]["replicates"]) > 1:

            jobfile.write("#BSUB -J " + jobname + "[1-" +
                          jobs[jobname]["replicates"] + "]\n")

    if jobs[jobname]["queue"] is not "":

        jobfile.write("#BSUB -q " + jobs[jobname]["queue"] + "\n")

    if jobs[jobname]["cluster"] is not "":

        jobfile.write("#BSUB -m " + jobs[jobname]["cluster"] + "\n")

    # Account to charge (if supplied).
    if hosts[jobs[jobname]["resource"]]["account"] is not "":

        # if no accountflag is provided use the default
        if hosts[jobs[jobname]["resource"]]["accountflag"] is "":

            jobfile.write("#BSUB -P " +
                          hosts[jobs[jobname]["resource"]]["account"] +
                          "\n")
        else:

            jobfile.write("#BSUB " +
                          hosts[jobs[jobname]["resource"]]["accountflag"] +
                          " " + hosts[jobs[jobname]["resource"]]["account"] +
                          "\n")

    jobfile.write("#BSUB -W " + jobs[jobname]["maxtime"] + "\n")

    jobfile.write("#BSUB -n " + jobs[jobname]["cores"] +
                  "\n")

    if jobs[jobname]["modules"] is not "":

        for module in jobs[jobname]["modules"].split(","):

            module = module.replace(" ", "")
            jobfile.write("\n" + "module load {0}\n\n" .format(module))

    mpirun = hosts[jobs[jobname]["resource"]]["handler"]

    # Single job
    if int(jobs[jobname]["replicates"]) == 1:

        jobfile.write(mpirun + " -lsf " + jobs[jobname]["executableargs"] +
                      "\n")

    # Job array
    elif int(jobs[jobname]["replicates"]) > 1:

        jobfile.write("cd rep${LSB_JOBINDEX}/\n" +
                      mpirun + " -lsf " + jobs[jobname]["executableargs"] +
                      "\n")

    # Close the file (housekeeping)
    jobfile.close()

    # Append lsf file to list of files ready for staging.
    jobs[jobname]["upload-include"] = (
        jobs[jobname]["upload-include"] + ", submit.lsf")
    jobs[jobname]["subfile"] = "submit.lsf"


def status(host, jobid):

    """Method for querying job."""

    state = ""

    try:

        shellout = SHELLWRAPPERS.sendtossh(host, ["bjobs -u " + host["user"] +
                                                  " | grep " + jobid])

        stat = shellout[0].split()

        if stat[2] == "PSUSP" or stat[2] == "USUSP" or stat[2] == "SSUSP":

            state = "Held"

        elif stat[2] == "PEND":

            state = "Queued"

        elif stat[2] == "RUN":

            state = "Running"

    except EX.SSHError:

        state = "Finished"

    return state


def submit(host, jobname, jobs):

    """Method for submitting job."""

    # Set the path to remoteworkdir/jobnamexxxxx
    path = jobs[jobname]["destdir"]

    # cd into the working directory and submit the job.
    cmd = ["cd " + path + "\n", "bsub < " +
           jobs[jobname]["subfile"] + "| grep -P -o '(?<=<)[0-9]*(?=>)'"]

    # Process the submit
    try:

        shellout = SHELLWRAPPERS.sendtossh(host, cmd)[0]

    except EX.SSHError:

        raise EX.JobsubmitError("  Something went wrong when submitting.")

    output = shellout.splitlines()[0]

    LOGGER.info("Job '{0}' submitted with id '{1}'" .format(jobname, output))

    jobs[jobname]["jobid"] = output



# Longbow is Copyright (C) of James T Gebbie-Rayet and Gareth B Shannon 2015.
#
# This file is part of the Longbow software which was developed as part of
# the HECBioSim project (http://www.hecbiosim.ac.uk/).
#
# HECBioSim facilitates and supports high-end computing within the
# UK biomolecular simulation community on resources such as ARCHER.
#
# Longbow is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Longbow is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Longbow.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains generic methods for preparing, submitting, deleting and
monitoring jobs. The methods contained within this module are all based on
generic job concepts. The specific functionality that comes from each scheduler
is accessed through the plug-in framework. To make use of these methods, the
plug-in framework must be present alongside the core library.

testenv()
    This method makes an attempt to test the environment and determine from
    a pre-configured list what scheduler and job submission handler is present
    on the machine.

delete()
    A method containing the generic and boiler plate Longbow code for deleting
    a job.

monitor()
    A method containing the generic and boiler plate Longbow code for
    monitoring a job, this method contains the entire structure of the loop
    that deals with monitoring jobs.

prepare()
    A method containing the generic and boiler plate Longbow code for
    constructing the submit file.

submit()
    A method containing the generic and boiler plate Longbow code for
    submitting a job.
"""

import time
import logging

# Depending on how longbow is installed/utilised the import will be slightly
# different, this should handle both cases.
try:

    CONFIGURATION = __import__("corelibs.configuration", fromlist=[''])
    EX = __import__("corelibs.exceptions", fromlist=[''])
    SCHEDULERS = __import__("plugins.schedulers", fromlist=[''])
    SHELLWRAPPERS = __import__("corelibs.shellwrappers", fromlist=[''])
    STAGING = __import__("corelibs.staging", fromlist=[''])

except ImportError:

    CONFIGURATION = __import__("Longbow.corelibs.configuration", fromlist=[''])
    EX = __import__("Longbow.corelibs.exceptions", fromlist=[''])
    SCHEDULERS = __import__("Longbow.plugins.schedulers", fromlist=[''])
    SHELLWRAPPERS = __import__("Longbow.corelibs.shellwrappers", fromlist=[''])
    STAGING = __import__("Longbow.corelibs.staging", fromlist=[''])

LOGGER = logging.getLogger("Longbow")


def testenv(hostconf, hosts, jobs):

    """
    This method makes an attempt to test the environment and determine from
    a pre-configured list what scheduler and job submission handler is present
    on the machine.

    Required arguments are:

    hostconf (string) - The path to the host configuration file, this should be
                        provided so that if any changes are made that they can
                        be saved.

    hosts (dictionary) - The Longbow hosts data structure, see configuration.py
                         for more information about the format of this
                         structure.

    jobs (dictionary) - The Longbow jobs data structure, see configuration.py
                        for more information about the format of this
                        structure.
    """

    schedulerqueries = getattr(SCHEDULERS, "QUERY")

    handlers = {
        "aprun": ["which aprun"],
        "mpirun": ["which mpirun"]
    }

    save = False

    checked = []

    # Take a look at each job.
    for job in jobs:

        resource = jobs[job]["resource"]

        # If we have not checked this host already
        if resource not in checked:

            # Make sure we don't check the same thing again.
            checked.extend([resource])

            # If we have no scheduler defined by the user then find it.
            if hosts[resource]["scheduler"] is "":

                LOGGER.info(
                    "No environment for this host '{0}' is specified - "
                    "attempting to determine it!".format(resource))

                # Go through the schedulers we are supporting.
                for param in schedulerqueries:

                    try:

                        SHELLWRAPPERS.sendtossh(hosts[resource],
                                                schedulerqueries[param])

                        hosts[resource]["scheduler"] = param

                        LOGGER.info(
                            "The environment on this host is '{0}'"
                            .format(param))
                        break

                    except EX.SSHError:

                        LOGGER.debug("Environment is not '{0}'".format(param))

                if hosts[resource]["scheduler"] is "":

                    raise EX.SchedulercheckError(
                        "Could not find the job scheduling system.")

                # If we changed anything then mark for saving.
                save = True

            else:

                LOGGER.info(
                    "The environment on host '{0}' is '{1}'"
                    .format(resource, hosts[resource]["scheduler"]))

            # If we have no job handler defined by the user then find it.
            if hosts[resource]["handler"] is "":

                LOGGER.info(
                    "No queue handler was specified for host '{0}' - "
                    "attempting to find it".format(resource))

                # Go through the handlers and find out which is there.
                # Load modules first as this is necessary for some remote
                # resources
                cmdmod = []

                for module in jobs[job]["modules"].split(","):

                    module = module.replace(" ", "")
                    cmdmod.extend(["module load " + module + "\n"])

                for param in handlers:

                    try:

                        cmd = cmdmod[:]
                        cmd.extend(handlers[param])
                        SHELLWRAPPERS.sendtossh(hosts[resource], cmd)

                        hosts[resource]["handler"] = param

                        LOGGER.info("The batch queue handler is '{0}'"
                                    .format(param))

                        break

                    except EX.SSHError:

                        LOGGER.debug(
                            "The batch queue handler is not '{0}'"
                            .format(param))

                if hosts[resource]["handler"] is "":

                    raise EX.HandlercheckError(
                        "Could not find the batch queue handler.")

                # If we changed anything then mark for saving.
                save = True

            else:

                LOGGER.info(
                    "The handler on host '{0}' is '{1}'"
                    .format(resource, hosts[resource]["handler"]))

    # Do we have anything to change in the host file.
    if save is True:

        CONFIGURATION.saveconfigs(hostconf, hosts)


def delete(hosts, jobs, jobname):

    """
    A method containing the generic and boiler plate Longbow code for deleting
    a job.

    Required arguments are:

    hosts (dictionary) - The Longbow hosts data structure, see configuration.py
                         for more information about the format of this
                         structure.

    jobs (dictionary) - The Longbow jobs data structure, see configuration.py
                        for more information about the format of this
                        structure.

    jobname (string) - The jobname of the job for deletion.
    """

    scheduler = hosts[jobs[jobname]["resource"]]["scheduler"]
    host = hosts[jobs[jobname]["resource"]]
    job = jobs[jobname]

    try:

        getattr(SCHEDULERS, scheduler.lower()).delete(host, job)

    except AttributeError:

        raise EX.PluginattributeError(
            "delete method cannot be found in plugin '{0}'"
            .format(scheduler))

    except EX.JobdeleteError:

        LOGGER.info("Unable to delete job '{0}'".format(jobname))


def monitor(hosts, jobs):

    """
    A method containing the generic and boiler plate Longbow code for
    monitoring a job, this method contains the entire structure of the loop
    that deals with monitoring jobs.

    Required arguments are:

    hosts (dictionary) - The Longbow hosts data structure, see configuration.py
                         for more information about the format of this
                         structure.

    jobs (dictionary) - The Longbow jobs data structure, see configuration.py
                        for more information about the format of this
                        structure.
    """

    LOGGER.info("Monitoring job/s")

    # Some initial values
    allfinished = False
    interval = 0

    # Find out which job has been set the highest polling frequency and use
    # that.
    for job in jobs:

        # Don't initialise if already set (submission errors are passed here)
        if "laststatus" not in jobs[job]:
            jobs[job]["laststatus"] = ""

        if interval < int(jobs[job]["frequency"]):

            interval = int(jobs[job]["frequency"])

    # Loop until all jobs are done.
    while allfinished is False:

        for job in jobs:

            if jobs[job]["laststatus"] != "Finished" and \
               jobs[job]["laststatus"] != "Submit Error":

                machine = jobs[job]["resource"]
                scheduler = hosts[machine]["scheduler"]
                host = hosts[machine]

                # Get the job status.
                try:

                    status = getattr(
                        SCHEDULERS, scheduler.lower()).status(
                            host, jobs[job]["jobid"])

                except AttributeError:

                    raise EX.PluginattributeError(
                        "status method cannot be found in plugin '{0}'"
                        .format(scheduler))

                # If the last status is different then change the flag (stops
                # logfile getting flooded!)
                if jobs[job]["laststatus"] != status:

                    jobs[job]["laststatus"] = status
                    LOGGER.info(
                        "Status of job '{0}' with id '{1}' is '{2}'"
                        .format(job, jobs[job]["jobid"], status))

                # If the job is not finished and we set the polling frequency
                # higher than 0 (off) then stage files.
                if (jobs[job]["laststatus"] == "Running" or
                        jobs[job]["laststatus"] == "Subjob(s) running" and
                        interval is not 0):

                    STAGING.stage_downstream(hosts, jobs, job)

                # If job is done wait 60 seconds then transfer files (this is
                # to stop users having to wait till all jobs end to grab last
                # bit of staged files.)
                if jobs[job]["laststatus"] == "Finished":

                    LOGGER.info(
                        "Job '{0}' is finishing, staging will begin in 60 "
                        "seconds".format(job))

                    time.sleep(60.0)

                    STAGING.stage_downstream(hosts, jobs, job)

        # If the polling interval is set at zero then staging will be disabled
        # however continue to poll jobs but do it on a low frequency. Staging
        # will however still occur once the job is finished.
        if interval is 0:

            time.sleep(120.0)

        # Find out if all jobs are completed.
        for job in jobs:

            if jobs[job]["laststatus"] != "Finished" and \
               jobs[job]["laststatus"] != "Submit Error":

                allfinished = False
                break

            allfinished = True

        # If we still have jobs running then wait here for desired time before
        # looping again.
        if allfinished is False:

            time.sleep(float(interval))

    LOGGER.info("All jobs are complete.")


def prepare(hosts, jobs):

    """
    A method containing the generic and boiler plate Longbow code for
    constructing the submit file.

    Required arguments are:

    hosts (dictionary) - The Longbow hosts data structure, see configuration.py
                         for more information about the format of this
                         structure.

    jobs (dictionary) - The Longbow jobs data structure, see configuration.py
                        for more information about the format of this
                        structure.
    """

    LOGGER.info("Creating submit files for job/s.")

    for job in jobs:

        scheduler = hosts[jobs[job]["resource"]]["scheduler"]

        try:

            getattr(SCHEDULERS, scheduler.lower()).prepare(hosts, job, jobs)

        except AttributeError:

            raise EX.PluginattributeError(
                "prepare method cannot be found in plugin '{0}'"
                .format(scheduler))

    LOGGER.info("Submit file/s created.")


def submit(hosts, jobs):

    """
    A method containing the generic and boiler plate Longbow code for
    submitting a job.

    Required arguments are:

    hosts (dictionary) - The Longbow hosts data structure, see configuration.py
                         for more information about the format of this
                         structure.

    jobs (dictionary) - The Longbow jobs data structure, see configuration.py
                        for more information about the format of this
                        structure.
    """

    LOGGER.info("Submitting job/s.")

    for job in jobs:

        machine = jobs[job]["resource"]
        scheduler = hosts[machine]["scheduler"]
        host = hosts[machine]

        try:

            getattr(SCHEDULERS, scheduler.lower()).submit(host, job, jobs)

        except AttributeError:

            raise EX.PluginattributeError(
                "submit method cannot be found in plugin '{0}'"
                .format(scheduler))

        except EX.JobsubmitError as err:

            LOGGER.error(
                "Submitting job '{0}' failed with message - '{1}'"
                .format(job, err))

            jobs[job]["laststatus"] = "Submit Error"

    LOGGER.info("Submission complete.")
